from __future__ import annotations

import asyncio
import base64
import json
import uuid
from itertools import chain
from typing import TYPE_CHECKING, Union

import boto3

from iambic.config.dynamic_config import ExtendsConfig, ExtendsConfigKey
from iambic.core.context import ctx
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import (
    BaseTemplate,
    ExecutionMessage,
    TemplateChangeDetails,
    Variable,
)
from iambic.core.parser import load_templates
from iambic.core.template_generation import get_existing_template_map
from iambic.core.utils import async_batch_processor, gather_templates, yaml
from iambic.plugins.v0_1_0.aws.event_bridge.models import (
    GroupMessageDetails,
    ManagedPolicyMessageDetails,
    PermissionSetMessageDetails,
    RoleMessageDetails,
    UserMessageDetails,
)
from iambic.plugins.v0_1_0.aws.iam.group.models import AWS_IAM_GROUP_TEMPLATE_TYPE
from iambic.plugins.v0_1_0.aws.iam.group.template_generation import (
    collect_aws_groups,
    generate_aws_group_templates,
)
from iambic.plugins.v0_1_0.aws.iam.policy.models import AWS_MANAGED_POLICY_TEMPLATE_TYPE
from iambic.plugins.v0_1_0.aws.iam.policy.template_generation import (
    collect_aws_managed_policies,
    generate_aws_managed_policy_templates,
)
from iambic.plugins.v0_1_0.aws.iam.role.template_generation import (
    collect_aws_roles,
    generate_aws_role_templates,
)
from iambic.plugins.v0_1_0.aws.iam.user.models import AWS_IAM_USER_TEMPLATE_TYPE
from iambic.plugins.v0_1_0.aws.iam.user.template_generation import (
    collect_aws_users,
    generate_aws_user_templates,
)
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.models import (
    AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE,
)
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.template_generation import (
    collect_aws_permission_sets,
    generate_aws_permission_set_templates,
)
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.utils import (
    generate_permission_set_map,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig


async def load(config: AWSConfig) -> AWSConfig:
    config_account_idx_map = {
        account.account_id: idx for idx, account in enumerate(config.accounts)
    }
    if config.organizations:
        if any(account.hub_role_arn for account in config.accounts):
            log.warning(
                "You have a `hub_role_arn` defined on an `aws.account` specified in your IAMbic configuration that will be ignored. "
                "IAMbic will prefer the `hub_role_arn` specified under your AWS Organization. To remove this message, "
                "please remove the `hub_role_arn` specified in an `AWS Account`."
            )
        orgs_accounts = await asyncio.gather(
            *[org.get_accounts() for org in config.organizations]
        )
        for org_accounts in orgs_accounts:
            for account in org_accounts:
                if (
                    account_elem := config_account_idx_map.get(account.account_id)
                ) is not None:
                    config.accounts[
                        account_elem
                    ].hub_session_info = account.hub_session_info
                    config.accounts[
                        account_elem
                    ].identity_center_details = account.identity_center_details
                else:
                    log.warning(
                        "Account not found in config. Account will be ignored.",
                        account_id=account.account_id,
                        account_name=account.account_name,
                    )
    elif config.accounts:
        hub_account = [account for account in config.accounts if account.hub_role_arn]
        if len(hub_account) > 1:
            raise AttributeError("Only one AWS Account can specify the hub_role_arn")
        elif not hub_account:
            raise AttributeError("One of the AWS Accounts must define the hub_role_arn")
        else:
            hub_account = hub_account[0]
            await hub_account.set_hub_session_info()
            hub_session_info = hub_account.hub_session_info
            if not hub_session_info:
                raise Exception("Unable to assume into the hub_role_arn")
            for account in config.accounts:
                if account.account_id != hub_account.account_id:
                    account.hub_session_info = hub_session_info

    # Set up the dynamic account variables
    for idx, account in enumerate(config.accounts):
        config.accounts[idx].variables.extend(
            [
                Variable(key="account_id", value=account.account_id),
                Variable(key="account_name", value=account.account_name),
            ]
        )

    # Preload the iam client to improve performance
    await asyncio.gather(
        *[account.get_boto3_client("iam") for account in config.accounts],
        return_exceptions=True,
    )

    return config


async def apply_identity_center_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    templates: list[BaseTemplate],
    remote_worker=None,
) -> list[TemplateChangeDetails]:
    """
    The async_apply_callable for IdentityCenter (SSO) resources.

    :param exe_message: Execution context
    :param config: The config object.
    :param templates: The list of templates to apply.
    :param remote_worker: The remote worker to use for applying templates.
    """
    await config.set_identity_center_details(exe_message.provider_id)
    return await async_batch_processor(
        [template.apply(config) for template in templates],
        5,
        0.5,
    )


async def apply_iam_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    templates: list[BaseTemplate],
    remote_worker=None,
) -> list[TemplateChangeDetails]:
    """
    The async_apply_callable for IAM resource.

    :param exe_message: Execution context
    :param config: The config object.
    :param templates: The list of templates to apply.
    :param remote_worker: The remote worker to use for applying templates.
    """
    if config.spoke_role_is_read_only:
        log.critical("Unable to apply resources when spoke_role_is_read_only is True")
        return []

    await generate_permission_set_map(config.accounts, templates)

    template_changes: list[TemplateChangeDetails] = []
    excluded_from_batch = [AWS_MANAGED_POLICY_TEMPLATE_TYPE]

    if managed_policy_tasks := [
        template.apply(config)
        for template in templates
        if template.template_type == AWS_MANAGED_POLICY_TEMPLATE_TYPE
    ]:
        # There are other template changes that could rely on the managed policy so create these first
        template_changes.extend(await async_batch_processor(managed_policy_tasks, 40))
        if len(template_changes) > len(managed_policy_tasks):
            # Give a few seconds to allow the managed policies to be created in AWS
            await asyncio.sleep(10)

    if any(
        template.template_type == AWS_IAM_GROUP_TEMPLATE_TYPE for template in templates
    ) and any(
        template.template_type == AWS_IAM_USER_TEMPLATE_TYPE for template in templates
    ):
        # There are user templates that may rely on the group so groups must be created first
        excluded_from_batch.append(AWS_IAM_GROUP_TEMPLATE_TYPE)
        group_tasks = [
            template.apply(config)
            for template in templates
            if template.template_type == AWS_IAM_GROUP_TEMPLATE_TYPE
        ]
        template_changes.extend(await async_batch_processor(group_tasks, 30))
        # Give a few seconds to allow the group to be created in AWS
        await asyncio.sleep(10)

    template_changes.extend(
        await async_batch_processor(
            [
                template.apply(config)
                for template in templates
                if template.template_type not in excluded_from_batch
            ],
            30,
        )
    )
    return template_changes


async def apply(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    templates: list[BaseTemplate],
    remote_worker=None,
) -> list[TemplateChangeDetails]:
    """
    The async_apply_callable for the AWS IambicPlugin class.

    :param exe_message: Execution context
    :param config: The config object.
    :param templates: The list of templates to apply.
    :param remote_worker: The remote worker to use for applying templates.
    """
    # TODO: Leverage exe_message as part of a distributed execution
    # TODO: Leverage remote_worker as part of a distributed execution

    identity_center_templates = []
    iam_templates = []
    tasks = []

    for template in templates:
        if template.template_type == AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE:
            identity_center_templates.append(template)
        else:
            iam_templates.append(template)

    if identity_center_templates:
        tasks.append(
            apply_identity_center_templates(
                exe_message, config, identity_center_templates, remote_worker
            )
        )

    if iam_templates:
        tasks.append(
            apply_iam_templates(exe_message, config, iam_templates, remote_worker)
        )

    template_changes = list(chain.from_iterable(await asyncio.gather(*tasks)))

    return [
        template_change
        for template_change in template_changes
        if template_change.proposed_changes or template_change.exceptions_seen
    ]


async def import_service_resources(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    service_name: str,
    async_collector_callables: list,
    async_generator_callables: list,
    messages: list = None,
    remote_worker=None,
    existing_template_map: dict = None,
):
    base_runner = bool(not exe_message.metadata)
    if not exe_message.metadata:
        exe_message = exe_message.copy()
        exe_message.metadata = dict(service=service_name)

    for async_collector_callable in async_collector_callables:
        tasks = []

        for account in config.accounts:
            task_message = exe_message.copy()

            if (
                task_message.provider_id
                and task_message.provider_id != account.account_id
            ):
                continue
            elif account.iambic_managed == IambicManaged.DISABLED:
                continue
            elif not task_message.provider_id:
                task_message.provider_id = account.account_id

            tasks.append(
                async_collector_callable(
                    task_message, config, existing_template_map, messages
                )
            )

        if tasks:
            if base_runner and ctx.use_remote and remote_worker and not messages:
                # TODO: Update to use the remote_worker
                await asyncio.gather(*tasks)
                # TODO: Add a process to gather status messages from the remote worker
            else:
                if remote_worker:
                    log.warning(
                        "The remote worker definition must be defined in the config to run remote execution."
                    )
                await asyncio.gather(*tasks)

    if base_runner:
        await asyncio.gather(
            *[
                async_generator_callable(
                    exe_message,
                    config,
                    base_output_dir,
                    existing_template_map,
                    messages,
                )
                for async_generator_callable in async_generator_callables
            ]
        )


async def import_identity_center_resources(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    messages: list = None,
    remote_worker=None,
    existing_template_map: dict = None,
):
    identity_center_config = config.copy()
    identity_center_config.accounts = [
        account
        for account in config.accounts
        if account.identity_center_details
        and account.iambic_managed != IambicManaged.DISABLED
    ]
    identity_center_accounts = [
        account.account_id for account in identity_center_config.accounts
    ]
    if not identity_center_accounts:
        return
    elif (
        exe_message.provider_id
        and exe_message.provider_id not in identity_center_accounts
    ):
        return

    await config.set_identity_center_details(exe_message.provider_id)
    await import_service_resources(
        exe_message,
        identity_center_config,
        base_output_dir,
        "identity_center",
        [collect_aws_permission_sets],
        [generate_aws_permission_set_templates],
        messages,
        remote_worker,
        existing_template_map,
    )


async def import_aws_resources(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    messages: list = None,
    remote_worker=None,
):
    tasks = []

    if not exe_message.metadata or exe_message.metadata["service"] == "identity_center":
        identity_center_template_map = None
        if not remote_worker or exe_message.metadata:
            identity_center_template_map = await get_existing_template_map(
                repo_dir=base_output_dir,
                template_type="AWS::IdentityCenter.*",
                nested=True,
            )

        tasks.append(
            import_identity_center_resources(
                exe_message,
                config,
                base_output_dir,
                messages,
                remote_worker,
                identity_center_template_map,
            )
        )

    if not exe_message.metadata or exe_message.metadata["service"] == "iam":
        iam_template_map = None

        if not remote_worker or exe_message.metadata:
            iam_template_map = await get_existing_template_map(
                repo_dir=base_output_dir,
                template_type="AWS::IAM.*",
                nested=True,
            )

        tasks.append(
            import_service_resources(
                exe_message,
                config,
                base_output_dir,
                "iam",
                [
                    collect_aws_roles,
                    collect_aws_groups,
                    collect_aws_users,
                    collect_aws_managed_policies,
                ],
                [
                    generate_aws_role_templates,
                    generate_aws_user_templates,
                    generate_aws_group_templates,
                    generate_aws_managed_policy_templates,
                ],
                messages,
                remote_worker,
                iam_template_map,
            )
        )

    await asyncio.gather(*tasks)


async def detect_changes(  # noqa: C901
    config: AWSConfig, repo_dir: str
) -> Union[str, None]:
    if not config.sqs_cloudtrail_changes_queues:
        log.debug("No cloudtrail changes queue arn found. Returning")
        return

    role_messages = []
    user_messages = []
    group_messages = []
    managed_policy_messages = []
    permission_set_messages = []
    commit_message = "Out of band changes detected.\nSummary:\n"

    for queue_arn in config.sqs_cloudtrail_changes_queues:
        queue_name = queue_arn.split(":")[-1]
        region_name = queue_arn.split(":")[3]
        session = await config.get_boto_session_from_arn(queue_arn, region_name)
        identity = session.client("sts").get_caller_identity()
        identity_arn_with_session_name = (
            identity["Arn"].replace(":sts:", ":iam:").replace("assumed-role", "role")
        )
        # TODO: This only works for same account identities. We need to do similar to NoqMeter,
        # check all roles we have access to on all accounts, or store this in configuration.
        # Then exclude these from the list of roles to check.

        identity_arn = "/".join(identity_arn_with_session_name.split("/")[0:2])
        sqs = session.client("sqs", region_name=region_name)
        queue_url_res = sqs.get_queue_url(QueueName=queue_name)
        queue_url = queue_url_res.get("QueueUrl")
        messages = sqs.receive_message(QueueUrl=queue_url, MaxNumberOfMessages=10).get(
            "Messages", []
        )

        while messages:
            processed_messages = []
            for message in messages:
                try:
                    processed_messages.append(
                        {
                            "Id": message["MessageId"],
                            "ReceiptHandle": message["ReceiptHandle"],
                        }
                    )
                    message_body = json.loads(message["Body"])
                    try:
                        if "Message" in message_body:
                            decoded_message = json.loads(message_body["Message"])[
                                "detail"
                            ]
                        else:
                            decoded_message = message_body["detail"]
                    except Exception as err:
                        log.debug(
                            "Unable to process message", error=str(err), message=message
                        )
                        processed_messages.append(
                            {
                                "Id": message["MessageId"],
                                "ReceiptHandle": message["ReceiptHandle"],
                            }
                        )
                        continue
                    actor = (
                        decoded_message.get("userIdentity", {})
                        .get("sessionContext", {})
                        .get("sessionIssuer", {})
                        .get("arn", "")
                    )
                    session_name = (
                        decoded_message.get("userIdentity", {})
                        .get("principalId")
                        .split(":")[-1]
                    )
                    if actor != identity_arn:
                        account_id = decoded_message.get("recipientAccountId")
                        request_params = decoded_message["requestParameters"]
                        event = decoded_message["eventName"]
                        resource_id = None
                        resource_type = None
                        if role_name := request_params.get("roleName"):
                            resource_id = role_name
                            resource_type = "Role"
                            role_messages.append(
                                RoleMessageDetails(
                                    account_id=account_id,
                                    role_name=role_name,
                                    delete=bool(event == "DeleteRole"),
                                )
                            )
                        elif user_name := request_params.get("userName"):
                            resource_id = user_name
                            resource_type = "User"
                            user_messages.append(
                                UserMessageDetails(
                                    account_id=account_id,
                                    role_name=user_name,
                                    delete=bool(event == "DeleteUser"),
                                )
                            )
                        elif group_name := request_params.get("groupName"):
                            resource_id = group_name
                            resource_type = "Group"
                            group_messages.append(
                                GroupMessageDetails(
                                    account_id=account_id,
                                    group_name=group_name,
                                    delete=bool(event == "DeleteGroup"),
                                )
                            )
                        elif policy_arn := request_params.get("policyArn"):
                            split_policy = policy_arn.split("/")
                            policy_name = split_policy[-1]
                            policy_path = (
                                "/"
                                if len(split_policy) == 2
                                else f"/{'/'.join(split_policy[1:-1])}/"
                            )
                            resource_id = policy_name
                            resource_type = "ManagedPolicy"
                            managed_policy_messages.append(
                                ManagedPolicyMessageDetails(
                                    account_id=account_id,
                                    policy_name=policy_name,
                                    policy_path=policy_path,
                                    delete=bool(
                                        decoded_message["eventName"] == "DeletePolicy"
                                    ),
                                )
                            )
                        elif permission_set_arn := request_params.get(
                            "permissionSetArn"
                        ):
                            resource_id = permission_set_arn
                            resource_type = "PermissionSet"
                            permission_set_messages.append(
                                PermissionSetMessageDetails(
                                    account_id=account_id,
                                    instance_arn=request_params.get("instanceArn"),
                                    permission_set_arn=permission_set_arn,
                                )
                            )

                        if resource_id:
                            commit_message = (
                                f"{commit_message}User {session_name} performed action {event} "
                                f"on {resource_type}({resource_id}) on account {account_id}.\n"
                            )
                except Exception as err:
                    log.debug(
                        "Unable to process message", error=str(err), message=message
                    )
                    continue

            sqs.delete_message_batch(QueueUrl=queue_url, Entries=processed_messages)
            messages = sqs.receive_message(
                QueueUrl=queue_url, MaxNumberOfMessages=10
            ).get("Messages", [])

    exe_message = ExecutionMessage(
        execution_id=str(uuid.uuid4()), command=Command.IMPORT, provider_type="aws"
    )
    collect_tasks = []
    iam_template_map = None
    identity_center_template_map = None

    if role_messages or user_messages or group_messages or managed_policy_messages:
        iam_template_map = await get_existing_template_map(
            repo_dir=repo_dir,
            template_type="AWS::IAM.*",
            nested=True,
        )

    if permission_set_messages:
        identity_center_template_map = await get_existing_template_map(
            repo_dir=repo_dir,
            template_type="AWS::IdentityCenter.*",
            nested=True,
        )

    if role_messages:
        collect_tasks.append(
            collect_aws_roles(exe_message, config, iam_template_map, role_messages)
        )
    if user_messages:
        collect_tasks.append(
            collect_aws_users(exe_message, config, iam_template_map, user_messages)
        )
    if group_messages:
        collect_tasks.append(
            collect_aws_groups(exe_message, config, iam_template_map, group_messages)
        )
    if managed_policy_messages:
        collect_tasks.append(
            collect_aws_managed_policies(
                exe_message, config, iam_template_map, managed_policy_messages
            )
        )
    if permission_set_messages:
        collect_tasks.append(
            collect_aws_permission_sets(
                exe_message,
                config,
                identity_center_template_map,
                permission_set_messages,
            )
        )

    if collect_tasks:
        await asyncio.gather(*collect_tasks)

        tasks = []
        if role_messages:
            tasks.append(
                generate_aws_role_templates(
                    exe_message, config, repo_dir, iam_template_map, role_messages
                )
            )
        if user_messages:
            tasks.append(
                generate_aws_user_templates(
                    exe_message, config, repo_dir, iam_template_map, user_messages
                )
            )
        if group_messages:
            tasks.append(
                generate_aws_group_templates(
                    exe_message, config, repo_dir, iam_template_map, group_messages
                )
            )
        if managed_policy_messages:
            tasks.append(
                generate_aws_managed_policy_templates(
                    exe_message,
                    config,
                    repo_dir,
                    iam_template_map,
                    managed_policy_messages,
                )
            )
        if permission_set_messages:
            tasks.append(
                generate_aws_permission_set_templates(
                    exe_message,
                    config,
                    repo_dir,
                    identity_center_template_map,
                    permission_set_messages,
                )
            )
        await asyncio.gather(*tasks)

        return commit_message


async def decode_aws_secret(config: AWSConfig, extend: ExtendsConfig) -> dict:
    if extend.key.value != ExtendsConfigKey.AWS_SECRETS_MANAGER.value:
        return {}

    assume_role_arn = extend.assume_role_arn
    secret_arn = extend.value
    region_name = secret_arn.split(":")[3]
    secret_account_id = secret_arn.split(":")[4]
    aws_account_map = {account.account_id: account for account in config.accounts}
    session = None

    if aws_account := aws_account_map.get(secret_account_id):
        if assume_role_arn == aws_account.spoke_role_arn:
            session = await aws_account.get_boto3_session(region_name=region_name)

    if not session and (config.accounts or config.organizations):
        if config.organizations:
            boto3_session = await config.organizations[0].get_boto3_session()
        else:
            hub_account = [
                account for account in config.accounts if account.hub_role_arn
            ][0]
            boto3_session = await hub_account.get_boto3_session()

        secret_account = AWSAccount(
            account_id=secret_account_id,
            account_name="Secret_Account",
            spoke_role_arn=assume_role_arn,
            hub_session_info=dict(boto3_session=boto3_session),
            boto3_session_map={},
        )
        session = await secret_account.get_boto3_session(region_name=region_name)
    elif not session:
        session = boto3.Session(region_name=region_name)

    try:
        client = session.client(service_name="secretsmanager")
        get_secret_value_response = client.get_secret_value(SecretId=secret_arn)
    except Exception:
        log.exception(
            "Unable to retrieve the AWS secret using the provided assume_role_arn",
            assume_role_arn=assume_role_arn,
            secret_arn=extend.value,
        )
        raise

    if "SecretString" in get_secret_value_response:
        return_val = get_secret_value_response["SecretString"]
    else:
        return_val = base64.b64decode(get_secret_value_response["SecretBinary"])

    return yaml.load(return_val)


async def discover_new_aws_accounts(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    config_account_idx_map: dict[str, int],
    orgs_accounts: list[list[AWSAccount]],
    repo_dir: str,
    remote_worker=None,
) -> bool:
    run_import = False
    accounts_to_apply = []
    for org_accounts in orgs_accounts:
        for account in org_accounts:
            if config_account_idx_map.get(account.account_id) is None:
                config.accounts.append(account)
                log.warning(
                    "New AWS account discovered. Adding account to config.",
                    account_id=account.account_id,
                    account_name=account.account_name,
                )
                if account.iambic_managed not in [
                    IambicManaged.DISABLED,
                    IambicManaged.IMPORT_ONLY,
                ]:
                    accounts_to_apply.append(account)

                if account.iambic_managed != IambicManaged.DISABLED:
                    run_import = True

    if accounts_to_apply:
        log.warning(
            "Applying templates to provision identities to the discovered account(s).",
        )
        templates = await gather_templates(repo_dir, "AWS.*")
        sub_message = exe_message.copy()
        sub_message.command = Command.APPLY
        sub_config = config.copy()
        sub_config.accounts = accounts_to_apply
        await apply(exe_message, sub_config, load_templates(templates), remote_worker)

    return run_import


async def discover_aws_account_attribute_changes(
    config: AWSConfig,
    config_account_idx_map: dict[str, int],
    orgs_accounts: list[list[AWSAccount]],
) -> bool:
    run_import = False
    for org_accounts in orgs_accounts:
        for account in org_accounts:
            if (
                account_elem := config_account_idx_map.get(account.account_id)
            ) is not None:
                config_account = config.accounts[account_elem]
                config_account_var_map = {
                    var["key"]: {"elem": idx, "value": var["value"]}
                    for idx, var in enumerate(config_account.variables)
                }

                if config_account.account_name != account.account_name:
                    log.warning(
                        "Updated AWS account name discovered. Updating account in config.",
                        account_id=account.account_id,
                        account_name=account.account_name,
                    )
                    config.accounts[account_elem].account_name = account.account_name
                    if account.iambic_managed != IambicManaged.DISABLED:
                        run_import = True

                for org_account_var in account.variables:
                    if config_account_var := config_account_var_map.get(
                        org_account_var.key
                    ):
                        if config_account_var["value"] != org_account_var.value:
                            log.warning(
                                "Mismatched variable on AWS account. Updating in config.",
                                account_id=account.account_id,
                                account_name=account.account_name,
                                variable_key=org_account_var.key,
                                discovered_value=org_account_var.value,
                                config_value=config_account_var["value"],
                            )
                            config.accounts[account_elem].variables[
                                config_account_var["elem"]
                            ].value = org_account_var.value
                            if account.iambic_managed != IambicManaged.DISABLED:
                                run_import = True

    return run_import


async def aws_account_update_and_discovery(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    repo_dir: str,
    remote_worker=None,
):
    """
    Update and discover AWS accounts.

    This function updates the list of AWS accounts in the `config` object by retrieving the list of accounts
    from the AWS Organizations and checking for any new or updated accounts. If any new or updated accounts
    are found, the function imports the AWS resources and regenerates the AWS templates.

    Args:
    - config (AWSConfig): The AWS configuration object.
    - repo_dir (str): The directory path for the repository.

    Returns:
    - None
    """
    # TODO: Leverage exe_message as part of a distributed execution
    # TODO: Leverage remote_worker as part of a distributed execution

    if not config.organizations:
        return

    ctx.eval_only = False
    config_account_idx_map = {
        account.account_id: idx for idx, account in enumerate(config.accounts)
    }

    orgs_accounts = await asyncio.gather(
        *[org.get_accounts() for org in config.organizations]
    )
    import_new_account = await discover_new_aws_accounts(
        exe_message,
        config,
        config_account_idx_map,
        orgs_accounts,
        repo_dir,
        remote_worker,
    )
    import_updated_account = await discover_aws_account_attribute_changes(
        config, config_account_idx_map, orgs_accounts
    )
    if import_new_account or import_updated_account:
        log.warning(
            "Running import to regenerate AWS templates.",
        )
        sub_message = exe_message.copy()
        sub_message.command = Command.IMPORT
        await import_aws_resources(sub_message, config, repo_dir, remote_worker)



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, List, Optional

import okta.models as models
from okta.models.user_status import UserStatus as OktaUserStatus

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import GlobalRetryController
from iambic.plugins.v0_1_0.okta.models import Group, User
from iambic.plugins.v0_1_0.okta.utils import generate_user_profile, handle_okta_fn

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.group.models import UserSimple
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaOrganization


async def list_all_users(okta_organization: OktaOrganization) -> List[User]:
    """
    List all users in Okta.

    Args:
    - okta_organization: An instance of the OktaOrganization class, which provides access to the Okta API.

    Returns:
    - A list of `User` instances, representing the users in Okta.
    """

    client = await okta_organization.get_okta_client()

    filter_operator = " OR ".join(
        [
            f'status eq "{status}"'
            for status in list(map(lambda s: s.name, list(OktaUserStatus)))
        ]
    )

    async with GlobalRetryController(
        fn_identifier="okta.list_users"
    ) as retry_controller:
        fn = functools.partial(
            client.list_users, query_params=dict(filter=filter_operator)
        )
        users, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            log.error("Error encountered when listing users", error=str(err))
            raise Exception(f"Error listing users: {str(err)}")

    while resp.has_next():
        async with GlobalRetryController(
            fn_identifier="okta.list_users"
        ) as retry_controller:
            next_users, err = await retry_controller(handle_okta_fn, resp.next)
        if err:
            log.error("Error encountered when listing users", error=str(err))
            raise Exception(f"Error listing users: {str(err)}")
        users.append(next_users)

    if not users:
        return []

    users_to_return = []
    for user in users:
        users_to_return.append(
            User(
                user_id=user.id,
                idp_name=okta_organization.idp_name,
                username=user.profile.login,
                status=user.status.value.lower(),
                extra=dict(
                    created=user.created,
                ),
                profile=await generate_user_profile(user),
            )
        )
    return users_to_return


async def list_group_users(group: Group, okta_organization: OktaOrganization) -> Group:
    """
    List the members of a group in Okta.

    Args:
    - group: An instance of the `Group` class, representing the group whose members we want to list.
    - okta_organization: An instance of the OktaOrganization class, which provides access to the Okta API.

    Returns:
    - The same instance of the `Group` class, with the `members` attribute populated with a list of `User` instances,
    representing the members of the group.
    """

    client = await okta_organization.get_okta_client()
    async with GlobalRetryController(
        fn_identifier="okta.list_group_users"
    ) as retry_controller:
        fn = functools.partial(client.list_group_users, group.group_id)
        users, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            log.error("Error encountered when listing users in group", error=str(err))
            raise Exception(f"Error listing users for group: {group.name}, {str(err)}")

    while resp.has_next():
        async with GlobalRetryController(
            fn_identifier="okta.list_group_users"
        ) as retry_controller:
            next_users, err = await retry_controller(handle_okta_fn, resp.next)
        if err:
            log.error("Error encountered when listing users in group", error=str(err))
            raise Exception(f"Error listing users for group: {group.name}, {str(err)}")
        users.append(next_users)

    if not users:
        # if there is really no users, we need to update our local knowledge of membership
        group.members = []
        return group

    users_to_return = []
    for user in users:
        users_to_return.append(
            User(
                user_id=user.id,
                idp_name=okta_organization.idp_name,
                username=user.profile.login,
                status=user.status.value.lower(),
                extra=dict(
                    created=user.created,
                ),
                profile=await generate_user_profile(user),
            )
        )
    group.members = users_to_return
    return group


async def list_all_groups(okta_organization: OktaOrganization) -> List[Group]:
    """
    List all groups in Okta.

    Args:
    - okta_organization: An instance of the OktaOrganization class, which provides access to the Okta API.

    Returns:
    - A list of `Group` instances, representing the groups in Okta.
    """

    client = await okta_organization.get_okta_client()
    async with GlobalRetryController(
        fn_identifier="okta.list_groups"
    ) as retry_controller:
        fn = functools.partial(client.list_groups)
        groups, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            log.error("Error encountered when listing groups", error=str(err))
            raise Exception(f"Error listing groups: {str(err)}")

    while resp.has_next():
        async with GlobalRetryController(
            fn_identifier="okta.list_groups"
        ) as retry_controller:
            next_groups, err = await retry_controller(handle_okta_fn, resp.next)
        if err:
            log.error("Error encountered when listing groups", error=str(err))
            raise Exception(f"Error listing groups: {str(err)}")
        groups.append(next_groups)

    if not groups:
        log.info(
            "No groups found in Okta Organization",
            okta_organization=okta_organization.idp_name,
        )
        return []
    tasks = []
    for group_raw in groups:
        group = Group(
            idp_name=okta_organization.idp_name,
            name=group_raw.profile.name,
            description=group_raw.profile.description,
            group_id=group_raw.id,
            attributes=dict(),
            extra=dict(
                created=group_raw.created,
            ),
        )
        tasks.append(list_group_users(group, okta_organization))
    return list(await asyncio.gather(*tasks))


async def get_group(
    group_id: str, group_name: str, okta_organization: OktaOrganization
) -> Optional[Group]:
    """
    Get a group from Okta using the okta library.

    Args:
    - group_id: The ID of the group to get.
    - group_name: The name of the group to get.
    - okta_organization: An instance of the OktaOrganization class, which provides access to the Okta API.

    Returns:
    - An instance of the `Group` class, representing the retrieved group. If an error occurs, returns None.
    """
    # Get a group from Okta using the okta library
    client = await okta_organization.get_okta_client()
    group = None
    if group_id:
        async with GlobalRetryController(
            fn_identifier="okta.get_group"
        ) as retry_controller:
            fn = functools.partial(client.get_group, group_id)
            group, resp, err = await retry_controller(handle_okta_fn, fn)
    if not group:
        # Try to get group by name
        async with GlobalRetryController(
            fn_identifier="okta.list_groups"
        ) as retry_controller:
            fn = functools.partial(client.list_groups, query_params={"q": group_name})
            groups, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            log.error(
                "Error encountered when getting group by name",
                group_id=group_id,
                group_name=group_name,
                error=str(err),
            )
            return None
        matching_group = None
        for matching_group in groups:
            if matching_group.profile.name == group_name:
                group = matching_group
                break

        if not group:
            return None
    if err:
        log.error(
            "Error encountered when getting group",
            group_id=group_id,
            group_name=group_name,
            error=str(err),
        )
        return None

    group = Group(
        idp_name=okta_organization.idp_name,
        name=group.profile.name,
        description=group.profile.description,
        group_id=group.id,
        attributes=dict(),
        extra=dict(
            created=group.created,
        ),
    )
    group = await list_group_users(group, okta_organization)
    return group


async def create_group(
    group_name: str,
    idp_name: str,
    description: str,
    okta_organization: OktaOrganization,
) -> Optional[Group]:
    """
    Create a new group in Okta.

    Args:
        group_name (str): The name of the group to create.
        idp_name (str): The IDP name for the group.
        description (str): The description for the group.
        okta_organization (OktaOrganization): The Okta organization to create the group in.

    Returns:
        Group: The created Group object.
    """

    # TODO: Need ProposedChanges, support context.execute = False
    client = await okta_organization.get_okta_client()

    group_profile = models.GroupProfile(
        {
            "name": group_name,
            "description": description,
        }
    )

    # Create the group
    group_model = models.Group({"profile": group_profile})
    if ctx.execute:
        async with GlobalRetryController(
            fn_identifier="okta.create_group"
        ) as retry_controller:
            fn = functools.partial(client.create_group, group_model)
            group, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception("Error creating group")
        group = Group(
            idp_name=idp_name,
            name=group_name,
            description=description,
            group_id=group.id,
            attributes=dict(),
            extra=dict(
                created=group.created,
            ),
        )
        return group
    return None


async def update_group_name(
    group: Group,
    new_name: str,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the name of a group in Okta.

    Args:
        group (Group): The group to update the name of.
        new_name (str): The new name for the group.
        okta_organization (OktaOrganization): The Okta organization to update the group in.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """

    response: list[ProposedChange] = []
    if group.name == new_name:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=group.group_id,
            resource_type=group.resource_type,
            attribute="group_name",
            current_value=group.name,
            new_value=new_name,
        )
    )
    group_model = models.Group(
        {
            "profile": models.GroupProfile(
                {
                    "name": new_name,
                    "description": group.description,
                }
            )
        }
    )
    if ctx.execute:
        client = await okta_organization.get_okta_client()
        async with GlobalRetryController(
            fn_identifier="okta.update_group"
        ) as retry_controller:
            fn = functools.partial(client.update_group, group.group_id, group_model)
            updated_group, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception("Error updating group")
        Group(
            idp_name=okta_organization.idp_name,
            name=updated_group.profile.name,
            description=updated_group.profile.description,
            group_id=updated_group.id,
            attributes=dict(),
            extra=dict(
                created=updated_group.created,
            ),
        )
    return response


async def update_group_description(
    group: Group,
    new_description: str,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the description of a group in Okta.

    Args:
        group (Group): The group to update the description of.
        new_description (str): The new description for the group.
        okta_organization (OktaOrganization): The Okta organization to update the group in.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = []
    if group.description == new_description:
        return response
    client = await okta_organization.get_okta_client()
    group_model = models.Group(
        {
            "profile": models.GroupProfile(
                {
                    "name": group.name,
                    "description": new_description,
                }
            )
        }
    )
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.UPDATE,
            resource_id=group.group_id,
            resource_type=group.resource_type,
            attribute="description",
            change_summary={
                "current_description": group.description,
                "proposed_description": new_description,
            },
            current_value=group.description,
            new_value=new_description,
        )
    )
    if ctx.execute:
        async with GlobalRetryController(
            fn_identifier="okta.update_group"
        ) as retry_controller:
            fn = functools.partial(client.update_group, group.group_id, group_model)
            new_group, resp, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception("Error updating group")
        Group(
            idp_name=okta_organization.idp_name,
            name=new_group.profile.name,
            description=new_group.profile.description,
            group_id=new_group.id,
            attributes=dict(),
            extra=dict(
                created=new_group.created,
            ),
        )
    return response


async def update_group_members(
    group: Group,
    new_members: List[UserSimple],
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the members of a group in Okta.

    Args:
        group (Group): The group to update the members of.
        new_members (List[UserSimple]): The new members to add to the group.
        okta_organization (OktaOrganization): The Okta organization to update the group in.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """

    client = await okta_organization.get_okta_client()
    response = []
    current_user_usernames = [user.username for user in group.members]
    desired_user_usernames = [user.username for user in new_members]
    users_to_remove = [
        user for user in current_user_usernames if user not in desired_user_usernames
    ]

    users_to_add = [
        user for user in desired_user_usernames if user not in current_user_usernames
    ]

    if users_to_remove:
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_id=group.group_id,
                resource_type=group.resource_type,
                attribute="users",
                change_summary={"UsersToRemove": list(users_to_remove)},
                current_value=current_user_usernames,
                new_value=users_to_remove,
            )
        )

    if users_to_add:
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_id=group.group_id,
                resource_type=group.resource_type,
                attribute="users",
                change_summary={"UsersToAdd": list(users_to_add)},
                current_value=current_user_usernames,
                new_value=users_to_add,
            )
        )

    if ctx.execute:
        for user in users_to_remove:
            async with GlobalRetryController(
                fn_identifier="okta.get_user"
            ) as retry_controller:
                fn = functools.partial(client.get_user, user)
                user_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error("Error retrieving user", user=user, **log_params)
                continue
            async with GlobalRetryController(
                fn_identifier="okta.remove_user_from_group"
            ) as retry_controller:
                fn = functools.partial(
                    client.remove_user_from_group,
                    group.group_id,
                    user_okta.id,
                )
                _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error(
                    "Error removing user from group",
                    user=user,
                    group=group.name,
                    **log_params,
                )
                continue

        for user in users_to_add:
            async with GlobalRetryController(
                fn_identifier="okta.get_user"
            ) as retry_controller:
                fn = functools.partial(client.get_user, user)
                user_okta, _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error("Error retrieving user", user=user, **log_params)
                continue
            async with GlobalRetryController(
                fn_identifier="okta.add_user_to_group"
            ) as retry_controller:
                fn = functools.partial(
                    client.add_user_to_group, group.group_id, user_okta.id
                )
                _, err = await retry_controller(handle_okta_fn, fn)
            if err:
                log.error(
                    "Error adding user to group",
                    user=user,
                    group=group.name,
                    **log_params,
                )
                continue
    return response


async def maybe_delete_group(
    delete: bool,
    group: Group,
    okta_organization: OktaOrganization,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Delete a group in Okta.

    Args:
        group (Group): The group to delete.
        okta_organization (OktaOrganization): The Okta organization to delete the group from.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = []
    client = await okta_organization.get_okta_client()
    if not delete:
        return response
    response.append(
        ProposedChange(
            change_type=ProposedChangeType.DELETE,
            resource_id=group.group_id,
            resource_type=group.resource_type,
            attribute="group",
            change_summary={"group": group.name},
            current_value=group.name,
            new_value=None,
        )
    )
    if ctx.execute:
        async with GlobalRetryController(
            fn_identifier="okta.delete_group"
        ) as retry_controller:
            fn = functools.partial(client.delete_group, group.group_id)
            _, err = await retry_controller(handle_okta_fn, fn)
        if err:
            raise Exception("Error deleting group")
    return response



from __future__ import annotations

import asyncio
from itertools import chain
from typing import TYPE_CHECKING, Any, List, Optional

from pydantic import Field, validator

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
)
from iambic.plugins.v0_1_0.google_workspace.group.utils import (
    create_group,
    get_group,
    get_group_members,
    maybe_delete_group,
    update_group_description,
    update_group_domain,
    update_group_email,
    update_group_members,
    update_group_name,
)
from iambic.plugins.v0_1_0.google_workspace.models import (
    GoogleTemplate,
    GroupMemberRole,
    GroupMemberStatus,
    GroupMemberSubscription,
    GroupMemberType,
    WhoCanInvite,
    WhoCanJoin,
    WhoCanPostMessage,
    WhoCanViewGroup,
    WhoCanViewMembership,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.google_workspace.iambic_plugin import GoogleProject

GOOGLE_GROUP_TEMPLATE_TYPE = "NOQ::GoogleWorkspace::Group"


class GroupMember(BaseModel, ExpiryModel):
    email: str
    expand: bool = Field(
        False,
        description="Expand the group into the members of the group. This is useful for nested groups.",
    )
    role: GroupMemberRole = GroupMemberRole.MEMBER
    type: GroupMemberType = GroupMemberType.USER
    status: GroupMemberStatus = GroupMemberStatus.ACTIVE
    subscription: GroupMemberSubscription = GroupMemberSubscription.EACH_EMAIL

    @property
    def resource_type(self):
        return "google:group:member"

    @property
    def resource_id(self):
        return self.email


class GroupProperties(BaseModel):
    name: str
    domain: str
    email: str
    description: str
    welcome_message: Optional[str]
    members: List[GroupMember]
    who_can_invite: WhoCanInvite = "ALL_MANAGERS_CAN_INVITE"
    who_can_join: WhoCanJoin = "CAN_REQUEST_TO_JOIN"
    who_can_post_message: WhoCanPostMessage = "NONE_CAN_POST"
    who_can_view_group: WhoCanViewGroup = "ALL_MANAGERS_CAN_VIEW"
    who_can_view_membership: WhoCanViewMembership = "ALL_MANAGERS_CAN_VIEW"
    iambic_managed: IambicManaged = IambicManaged.UNDEFINED
    extra: Any = Field(None, description=("Extra attributes to store"))

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"extra", "metadata_commented_dict"}

    @property
    def resource_type(self):
        return "google:group:template"

    @property
    def resource_id(self):
        return self.email

    @validator("members")
    def sort_groups(cls, v: list[GroupMember]):
        sorted_v = sorted(v, key=lambda member: member.email)
        return sorted_v


class GoogleWorkspaceGroupTemplate(GoogleTemplate, ExpiryModel):
    template_type = GOOGLE_GROUP_TEMPLATE_TYPE
    owner: Optional[str] = Field(None, description="Owner of the group")
    properties: GroupProperties

    def apply_resource_dict(self, google_project: GoogleProject):
        return {
            "name": self.properties.name,
            "email": self.properties.email,
            "description": self.properties.description,
            "members": self.properties.members,
        }

    async def _apply_to_account(
        self, google_project: GoogleProject
    ) -> AccountChangeDetails:
        proposed_group = self.apply_resource_dict(google_project)
        change_details = AccountChangeDetails(
            account=self.properties.domain,
            resource_id=self.properties.email,
            resource_type=self.properties.resource_type,
            new_value=proposed_group,  # TODO fix
            proposed_changes=[],
        )

        log_params = dict(
            resource_type=self.properties.resource_type,
            resource_id=self.properties.email,
            account=str(self.properties.domain),
        )

        current_group = await get_group(
            self.properties.email, self.properties.domain, google_project
        )
        if current_group:
            change_details.current_value = current_group

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                change_details.new_value = {}
                return change_details

        group_exists = bool(current_group)

        tasks = []

        await self.remove_expired_resources()

        if not group_exists:
            if self.deleted:
                log.info(
                    "Resource is marked for deletion, but does not exist in the cloud. Skipping.",
                )
                return change_details
            change_details.extend_changes(
                [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_id=self.properties.email,
                        resource_type=self.properties.resource_type,
                    )
                ]
            )
            log_str = "New resource found in code."
            if not ctx.execute:
                log.info(log_str, **log_params)
                # Exit now because apply functions won't work if resource doesn't exist
                return change_details

            log_str = f"{log_str} Creating resource..."
            log.info(log_str, **log_params)

            await create_group(
                id=self.properties.email,
                domain=self.properties.domain,
                email=self.properties.email,
                name=self.properties.name,
                description=self.properties.description,
                google_project=google_project,
            )
            group_get_attempt = 0
            while not current_group and group_get_attempt < 5:
                # handle fetching group too fast
                group_get_attempt = group_get_attempt + 1
                current_group = await get_group(
                    self.properties.email, self.properties.domain, google_project
                )
                if not current_group:
                    await asyncio.sleep(1)

            if current_group:
                change_details.current_value = current_group

        # TODO: Support group expansion
        tasks.extend(
            [
                update_group_domain(
                    current_group.properties.domain,
                    self.properties.domain,
                    log_params,
                ),
                update_group_email(
                    current_group.properties.email,
                    self.properties.email,
                    self.properties.domain,
                    google_project,
                    log_params,
                ),
                update_group_name(
                    self.properties.email,
                    current_group.properties.name,
                    self.properties.name,
                    self.properties.domain,
                    google_project,
                    log_params,
                ),
                update_group_description(
                    self.properties.email,
                    current_group.properties.description,
                    self.properties.description,
                    self.properties.domain,
                    google_project,
                    log_params,
                ),
                update_group_members(
                    self.properties.email,
                    current_group.properties.members,
                    [
                        member
                        for member in self.properties.members
                        if not member.deleted
                    ],
                    self.properties.domain,
                    google_project,
                    log_params,
                ),
                maybe_delete_group(
                    self,
                    google_project,
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            change_details.extend_changes(list(chain.from_iterable(changes_made)))

        if ctx.execute:
            log.debug(
                "Successfully finished execution for resource",
                changes_made=bool(change_details.proposed_changes),
                **log_params,
            )
            if self.deleted:
                self.delete()
            self.write()
        else:
            log.debug(
                "Successfully finished scanning for drift for resource",
                requires_changes=bool(change_details.proposed_changes),
                **log_params,
            )

        return change_details

    @property
    def resource_type(self):
        return "google:group"

    @property
    def default_file_path(self):
        file_name = f"{self.properties.email.split('@')[0]}.yaml"
        return f"resources/google/groups/{self.properties.domain}/{file_name}"

    def _is_iambic_import_only(self, google_project: GoogleProject):
        return (
            google_project.iambic_managed == IambicManaged.IMPORT_ONLY
            or self.iambic_managed == IambicManaged.IMPORT_ONLY
        )


async def get_group_template(service, group, domain) -> GoogleWorkspaceGroupTemplate:
    members = await get_group_members(service, group)

    file_name = f"{group['email'].split('@')[0]}.yaml"
    return GoogleWorkspaceGroupTemplate(
        file_path=f"resources/google/groups/{domain}/{file_name}",
        properties=dict(
            domain=domain,
            name=group["name"],
            email=group["email"],
            description=group["description"],
            members=members,
        ),
    )



from __future__ import annotations

import asyncio
import datetime
import json
import os
import re
import shutil
import sys
import tempfile
import time
import traceback
import uuid
from enum import Enum
from typing import Any, Callable
from urllib.parse import urlparse

import github
from github.PullRequest import PullRequest

import iambic.output.markdown
from iambic.config.dynamic_config import load_config
from iambic.config.utils import resolve_config_template_path
from iambic.core.context import ctx
from iambic.core.git import Repo, clone_git_repo, get_remote_default_branch
from iambic.core.iambic_enum import Command
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage, TemplateChangeDetails
from iambic.core.utils import yaml
from iambic.main import run_apply, run_detect, run_expire, run_git_apply, run_git_plan

iambic_app = __import__("iambic.lambda.app", globals(), locals(), [], 0)
lambda_run_handler = getattr(iambic_app, "lambda").app.run_handler


MERGEABLE_STATE_CLEAN = "clean"
MERGEABLE_STATE_BLOCKED = "blocked"


COMMIT_MESSAGE_USER_NAME = "Iambic Automation"
COMMIT_MESSAGE_USER_EMAIL = "iambic-automation@iambic.org"
COMMIT_MESSAGE_FOR_DETECT = "Import changes from detect operation"
COMMIT_MESSAGE_FOR_IMPORT = "Import changes from import operation"
COMMIT_MESSAGE_FOR_EXPIRE = "Periodic Expiration"
COMMIT_MESSAGE_FOR_GIT_APPLY_ABSOLUTE_TIME = "Replace relative time with absolute time"
SHARED_CONTAINER_GITHUB_DIRECTORY = "/root/data"


def init_shared_data_directory():
    # Have to be careful when setting REPO_BASE_PATH because lambda execution
    # environment can be very straight on only /tmp is writable
    if os.environ.get("AWS_LAMBDA_FUNCTION_NAME", False):
        temp_templates_directory = tempfile.mkdtemp(prefix="github")
        SHARED_CONTAINER_GITHUB_DIRECTORY = f"{temp_templates_directory}/data/"
    else:
        SHARED_CONTAINER_GITHUB_DIRECTORY = "/root/data"
    this_module = sys.modules[__name__]
    setattr(
        this_module,
        "SHARED_CONTAINER_GITHUB_DIRECTORY",
        SHARED_CONTAINER_GITHUB_DIRECTORY,
    )


def get_lambda_repo_path() -> str:
    return getattr(iambic_app, "lambda").app.REPO_BASE_PATH


class HandleIssueCommentReturnCode(Enum):
    UNDEFINED = 1
    NO_MATCHING_BODY = 2
    MERGEABLE_STATE_NOT_CLEAN = 3
    MERGED = 4
    PLANNED = 5


# context is a dictionary structure published by Github Action
# https://docs.github.com/en/actions/learn-github-actions/contexts#github-context
def run_handler(context: dict[str, Any]):
    github_token: str = context["token"]
    event_name: str = context["event_name"]
    log_params = {"event_name": event_name}
    log.info("run_handler", **log_params)
    github_client = github.Github(github_token)

    getattr(iambic_app, "lambda").app.init_plan_output_path()
    getattr(iambic_app, "lambda").app.init_repo_base_path()

    # TODO Support Github Enterprise with custom hostname
    # g = Github(base_url="https://{hostname}/api/v3", login_or_token="access_token")

    f: Callable[[github.Github, dict[str, Any]], None] = EVENT_DISPATCH_MAP.get(
        event_name
    )
    if f:
        f(github_client, context)
    else:
        log.error("no supported handler")
        raise Exception("no supported handler")


def handle_iambic_command(
    github_client: github.Github, context: dict[str, Any]
) -> None:
    github_token: str = context["iambic"]["GH_OVERRIDE_TOKEN"]
    command: str = context["iambic"]["IAMBIC_CLOUD_IMPORT_CMD"]
    github_client = github.Github(github_token)
    f: Callable[
        [github.Github, dict[str, Any]], None
    ] = IAMBIC_CLOUD_IMPORT_DISPATCH_MAP.get(command)
    if f:
        f(github_client, context)
    else:
        log.error("no supported handler")
        raise Exception("no supported handler")


def format_github_url(repository_url: str, github_token: str) -> str:
    parse_result = urlparse(repository_url)
    return parse_result._replace(
        netloc="oauth2:{0}@{1}".format(github_token, parse_result.netloc)
    ).geturl()


def prepare_local_repo_for_new_commits(
    repo_url: str, repo_path: str, purpose: str
) -> Repo:
    if len(os.listdir(repo_path)) > 0:
        raise Exception(f"{repo_path} already exists. This is unexpected.")
    cloned_repo = clone_git_repo(repo_url, repo_path, None)

    repo_config_writer = cloned_repo.config_writer()
    repo_config_writer.set_value("user", "name", COMMIT_MESSAGE_USER_NAME)
    repo_config_writer.set_value("user", "email", COMMIT_MESSAGE_USER_EMAIL)
    repo_config_writer.release()

    default_branch = get_remote_default_branch(cloned_repo)
    cloned_repo.git.checkout("-b", f"attempt/{purpose}", default_branch)

    return cloned_repo


def is_last_commit_relative_to_absolute_change(
    repo_url: str, pull_request_branch_name: str
) -> Repo:
    repo_path = tempfile.mkdtemp(prefix="github")

    if len(os.listdir(repo_path)) > 0:
        raise Exception(f"{repo_path} already exists. This is unexpected.")
    cloned_repo = clone_git_repo(repo_url, repo_path, pull_request_branch_name)
    return cloned_repo.head.commit.message == COMMIT_MESSAGE_FOR_GIT_APPLY_ABSOLUTE_TIME


def prepare_local_repo(
    repo_url: str, repo_path: str, pull_request_branch_name: str
) -> Repo:
    if len(os.listdir(repo_path)) > 0:
        raise Exception(f"{repo_path} already exists. This is unexpected.")
    cloned_repo = clone_git_repo(repo_url, repo_path, None)
    for remote in cloned_repo.remotes:
        remote.fetch()
    default_branch = get_remote_default_branch(cloned_repo)
    cloned_repo.git.checkout("-b", "attempt/git-apply", default_branch)

    # Note, this is for local usage, we don't actually
    # forward this commit upstream
    repo_config_writer = cloned_repo.config_writer()
    repo_config_writer.set_value("user", "name", COMMIT_MESSAGE_USER_NAME)
    repo_config_writer.set_value("user", "email", COMMIT_MESSAGE_USER_EMAIL)
    repo_config_writer.release()

    cloned_repo.git.merge(f"origin/{pull_request_branch_name}")
    log_params = {"last_commit_message": cloned_repo.head.commit.message}
    log.info(cloned_repo.head.commit.message, **log_params)
    return cloned_repo


def load_proposed_changes(filepath: str) -> dict:
    if not os.path.exists(filepath):
        return {}
    with open(filepath, "r") as f:
        return yaml.load(f)


# TODO do more formatting to emphasize resources deletion
def format_proposed_changes(changes: dict) -> str:
    pass


TRUNCATED_WARNING = (
    "Plan is too large and truncated. View Run link to check the entire plan."
)
BODY_MAX_LENGTH = 65000
TRUNCATED_BODY_MAX_LENGTH = BODY_MAX_LENGTH - len(TRUNCATED_WARNING)
GIT_APPLY_COMMENT_TEMPLATE = """iambic {iambic_op} ran with:

```yaml
{plan}
```

<a href="{run_url}">Run</a>
"""


def ensure_body_length_fits_github_spec(body: str, blob_html_url: str = None) -> str:
    if len(body) > BODY_MAX_LENGTH:
        body = TRUNCATED_WARNING + f"""<a href="{blob_html_url}">Run</a>"""
    return body


def post_result_as_pr_comment(
    pull_request: PullRequest,
    context: dict[str, Any],
    iambic_op: str,
    proposed_changes_path: str,
) -> None:
    if context:
        run_url = (
            context["server_url"]
            + "/"
            + context["repository"]
            + "/actions/runs/"
            + context["run_id"]
            + "/attempts/"
            + context["run_attempt"]
        )
    else:
        run_url = "lambda implementation not currently supported run_url"  # FIXME
    lines = []
    if not proposed_changes_path:
        cwd = os.getcwd()
        proposed_changes_path = f"{cwd}/proposed_changes.yaml"
    if os.path.exists(proposed_changes_path):
        with open(proposed_changes_path) as f:
            lines = f.readlines()
    plan = "".join(lines) if lines else "no changes"
    body = GIT_APPLY_COMMENT_TEMPLATE.format(
        plan=plan, run_url=run_url, iambic_op=iambic_op
    )
    _post_render_content_as_pr_comment(body)


def _post_render_content_as_pr_comment(
    pull_request: PullRequest,
    rendered_content: str,
    blob_html_url: str = None,
) -> None:
    body = ensure_body_length_fits_github_spec(
        rendered_content, blob_html_url=blob_html_url
    )
    pull_request.create_issue_comment(body)


def copy_data_to_data_directory() -> None:
    cwd = os.getcwd()
    filepath = f"{cwd}/proposed_changes.yaml"
    dest_dir = SHARED_CONTAINER_GITHUB_DIRECTORY
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    if os.path.exists(filepath):
        shutil.copy(filepath, f"{dest_dir}/proposed_changes.yaml")


IAMBIC_SESSION_NAME_TEMPLATE = "org={org},repo={repo},pr={number}"
SESSION_NAME_REGEX = re.compile("[\\w=,.@-]+")


def get_session_name(repo_name: str, pull_request_number: str) -> str:
    org = ""
    repo = ""
    repo_parts = repo_name.split("/")
    if len(repo_parts) == 2:
        repo = repo_parts[1]
        org = repo_parts[0]
    else:
        repo = repo_name
    pending_session_name = IAMBIC_SESSION_NAME_TEMPLATE.format(
        org=org, repo=repo, number=pull_request_number
    )
    session_name = "".join(
        [c for c in pending_session_name if SESSION_NAME_REGEX.match(c)]
    )
    if session_name != pending_session_name:
        log_params = {
            "pending_session_name": pending_session_name,
            "session_name": session_name,
        }
        log.error("lossy-session-name", **log_params)
    return session_name


def handle_issue_comment(
    github_client: github.Github, context: dict[str, Any]
) -> HandleIssueCommentReturnCode:
    comment_body = context["event"]["comment"]["body"]
    log_params = {"COMMENT_DISPATCH_MAP_KEYS": COMMENT_DISPATCH_MAP.keys()}
    log.info("COMMENT_DISPATCH_MAP keys", **log_params)
    if comment_body not in COMMENT_DISPATCH_MAP:
        log_params = {"comment_body": comment_body}
        log.error("handle_issue_comment: no op", **log_params)
        return HandleIssueCommentReturnCode.NO_MATCHING_BODY

    github_token = context["token"]
    # repo_name is already in the format {repo_owner}/{repo_short_name}
    repo_name = context["repository"]
    pull_number = context["event"]["issue"]["number"]
    repository_url = context["event"]["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)
    # repository_url_token
    templates_repo = github_client.get_repo(repo_name)
    pull_request = templates_repo.get_pull(pull_number)
    pull_request_branch_name = pull_request.head.ref
    log_params = {"pull_request_branch_name": pull_request_branch_name}
    log.info("PR remote branch name", **log_params)

    comment_func: Callable = COMMENT_DISPATCH_MAP[comment_body]
    return comment_func(
        context,
        github_client,
        templates_repo,
        pull_request,
        repo_name,
        pull_number,
        pull_request_branch_name,
        repo_url,
    )


def handle_iambic_git_apply(
    context: dict[str, Any],
    github_client: github.Github,
    templates_repo: github.Repo,
    pull_request: PullRequest,
    repo_name: str,
    pull_number: str,
    pull_request_branch_name: str,
    repo_url: str,
    proposed_changes_path: str = None,
):
    if pull_request.mergeable_state != MERGEABLE_STATE_CLEAN:
        # TODO log error and also make a comment to PR
        pull_request.create_issue_comment(
            "Mergable state is {0}. This probably means that the necessary approvals have not been granted for the request.".format(
                pull_request.mergeable_state
            )
        )
        return HandleIssueCommentReturnCode.MERGEABLE_STATE_NOT_CLEAN

    session_name = get_session_name(repo_name, pull_number)
    os.environ["IAMBIC_SESSION_NAME"] = session_name

    try:
        # merge_sha is used when we trigger a merge
        merge_sha = pull_request.head.sha

        repo_dir = get_lambda_repo_path()
        repo = prepare_local_repo(repo_url, repo_dir, pull_request_branch_name)
        # local_sha_before_git_apply may not match the initial merge
        # sha because we apply the PR to local checkout tracking branch
        local_sha_before_git_apply = repo.head.commit.hexsha

        if proposed_changes_path:
            # code smell to have to change a module variable
            # to control the destination of proposed_changes.yaml
            # It's questionable if we still need to depend on the lambda interface
            # because lambda interface was created to dynamic populate template config
            # but templates config is now already stored in the templates repo itself.
            getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH = proposed_changes_path

        template_changes = run_git_apply(
            False, None, None, repo_dir, proposed_changes_path
        )

        # In the event git_apply changes relative time to absolute time
        repo.git.add(".")
        diff_list = repo.head.commit.diff()
        if len(diff_list) > 0:
            repo.git.commit("-m", COMMIT_MESSAGE_FOR_GIT_APPLY_ABSOLUTE_TIME)
        else:
            log.debug("git_apply did not introduce additional changes")

        local_sha_after_git_apply = repo.head.commit.hexsha
        if local_sha_before_git_apply != local_sha_after_git_apply:
            # signal changes due to git-apply
            repo.remotes.origin.push(
                refspec=f"HEAD:{pull_request_branch_name}"
            ).raise_if_error()
            log_params = {"sha": local_sha_after_git_apply}
            log.info("git-apply new sha is", **log_params)
            # update merge sha because we add new commits to pull request
            merge_sha = local_sha_after_git_apply

        _process_template_changes(
            github_client,
            templates_repo,
            pull_request,
            pull_number,
            proposed_changes_path,
            template_changes,
            "apply",
        )
        copy_data_to_data_directory()

        maybe_merge(templates_repo, pull_number, merge_sha)
        return HandleIssueCommentReturnCode.MERGED

    except Exception as e:
        captured_traceback = traceback.format_exc()
        log.error("fault", exception=captured_traceback)
        pull_request.create_issue_comment(
            "exception during apply is {0} \n ```{1}```".format(
                pull_request.mergeable_state, captured_traceback
            )
        )
        raise e


def maybe_merge(
    templates_repo: github.Repo,
    # pull_request: PullRequest,
    pull_number: int,
    merge_sha: str,
    max_attempts: int = 5,
    sleep_interval: float = 5,
):
    """Attempts to merge the PR at specific sha, this function will retry a few times because
    desired sha maybe not available yet"""
    attempts = 0
    merge_status = None
    last_known_traceback = None
    while attempts < max_attempts:
        pull_request = templates_repo.get_pull(pull_number)
        try:
            merge_status = pull_request.merge(sha=merge_sha)
            break
        except github.GithubException:
            last_known_traceback = traceback.format_exc()
            attempts += 1
            time.sleep(sleep_interval)
    if merge_status is None:
        raise RuntimeError(
            f"Fail to merge PR. Target sha is {merge_sha}. last_known_trace_back is {last_known_traceback}"
        )


def _post_artifact_to_companion_repository(
    github_client: github.Github,
    templates_repo: github.Repo,
    pull_number: str,
    op_name: str,
    proposed_changes_path: str,
    markdown_summary: str,
    default_base_name: str = "proposed_changes.yaml",
):
    url = None
    try:
        lines = []
        if os.path.exists(proposed_changes_path):
            with open(proposed_changes_path) as f:
                lines = f.readlines()
            gist_repo_name = f"{templates_repo.full_name}-gist"
            gist_repo = github_client.get_repo(gist_repo_name)
            now_timestamp = datetime.datetime.now()
            pr_prefix = f"pr-{pull_number}" if pull_number else "no-pr"
            yaml_repo_path = (
                f"{pr_prefix}/{op_name}/{now_timestamp}/{default_base_name}"
            )
            md_repo_path = f"{pr_prefix}/{op_name}/{now_timestamp}/summary.md"
            gist_repo.create_file(yaml_repo_path, op_name, "".join(lines))
            result = gist_repo.create_file(md_repo_path, op_name, markdown_summary)
            url = result["content"].html_url
    except Exception:
        # Decision to keep going is we do not want a failure of posting to companion
        # repository (for full machine and human readable contents) to stop the rest of
        # plan and apply operations. Companion repository is a separate failure domain.
        captured_traceback = traceback.format_exc()
        log.error("fault", exception=captured_traceback)
    return url


def handle_iambic_git_plan(
    context: dict[str, Any],
    github_client: github.Github,
    templates_repo: github.Repo,
    pull_request: PullRequest,
    repo_name: str,
    pull_number: str,
    pull_request_branch_name: str,
    repo_url: str,
    proposed_changes_path: str = None,
):
    session_name = get_session_name(repo_name, pull_number)
    os.environ["IAMBIC_SESSION_NAME"] = session_name

    try:
        repo_dir = get_lambda_repo_path()
        prepare_local_repo(repo_url, repo_dir, pull_request_branch_name)

        if proposed_changes_path:
            # code smell to have to change a module variable
            # to control the destination of proposed_changes.yaml
            # It's questionable if we still need to depend on the lambda interface
            # because lambda interface was created to dynamic populate template config
            # but templates config is now already stored in the templates repo itself.
            getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH = proposed_changes_path

        template_changes = run_git_plan(proposed_changes_path, repo_dir)
        _process_template_changes(
            github_client,
            templates_repo,
            pull_request,
            pull_number,
            proposed_changes_path,
            template_changes,
            "plan",
        )
        copy_data_to_data_directory()
        return HandleIssueCommentReturnCode.PLANNED
    except Exception as e:
        captured_traceback = traceback.format_exc()
        log.error("fault", exception=captured_traceback)
        pull_request.create_issue_comment(
            "exception during plan is {0} \n ```{1}```".format(
                pull_request.mergeable_state, captured_traceback
            )
        )
        raise e


def github_app_workflow_wrapper(workflow_func: Callable, ux_op_name: str) -> Callable:
    def wrapped_workflow_func(
        github_client: github.Github,
        templates_repo: github.Repo,
        repo_name: str,
        repo_url: str,
        default_branch: str,
        proposed_changes_path: str = None,
    ):
        pull_number = (
            0  # 0 is not a valid pull number in github. workflow implementation
        )
        # does not have an associated PR. This is the path of least resistance adaption
        # for stable session name
        session_name = get_session_name(repo_name, pull_number)
        os.environ["IAMBIC_SESSION_NAME"] = session_name

        try:
            if proposed_changes_path:
                # code smell to have to change a module variable
                # to control the destination of proposed_changes.yaml
                # It's questionable if we still need to depend on the lambda interface
                # because lambda interface was created to dynamic populate template config
                # but templates config is now already stored in the templates repo itself.
                getattr(
                    iambic_app, "lambda"
                ).app.PLAN_OUTPUT_PATH = proposed_changes_path

            template_changes = workflow_func(repo_url, default_branch)
            _process_template_changes(
                github_client,
                templates_repo,
                None,
                pull_number,
                proposed_changes_path,
                template_changes,
                f"{ux_op_name}",  # TODO this can probably be improved for user-experience
            )
        except Exception as e:
            captured_traceback = traceback.format_exc()
            log.error("fault", exception=captured_traceback)
            try:
                temp_dir = tempfile.mkdtemp(suffix=None, prefix=None, dir=None)
                with open(f"{temp_dir}/crash.txt", "w") as f:
                    f.write(captured_traceback)
                _post_artifact_to_companion_repository(
                    github_client,
                    templates_repo,
                    pull_number,
                    f"{ux_op_name}",
                    f"{temp_dir}/crash.txt",
                    captured_traceback,
                    default_base_name="crash.txt",
                )
            except Exception:
                captured_traceback = traceback.format_exc()
                log.error(
                    "fail to post exception to companion repo",
                    exception=captured_traceback,
                )
            finally:
                if temp_dir:
                    shutil.rmtree(temp_dir)
            raise e

    return wrapped_workflow_func


def _process_template_changes(
    github_client: github.Github,
    templates_repo: github.Repo,
    pull_request: PullRequest,
    pull_number: str,
    proposed_changes_path: str,  # Path where we are sourcing the machine output
    template_changes: list[TemplateChangeDetails],
    op_name: str,  # Examples are "plan", "apply"
):
    html_url = ""
    if template_changes and len(template_changes) > 0:
        rendered_content = iambic.output.markdown.gh_render_resource_changes(
            template_changes
        )
        html_url = _post_artifact_to_companion_repository(
            github_client,
            templates_repo,
            pull_number,
            op_name,
            proposed_changes_path,
            rendered_content,
        )
    else:
        rendered_content = "no changes detected"

    rendered_content = f"""Reacting to `{op_name}`\n\n{rendered_content}\n\n <a href="{html_url}">Run</a>"""
    if pull_request:
        _post_render_content_as_pr_comment(
            pull_request, rendered_content, blob_html_url=html_url
        )


def handle_pull_request(github_client: github.Github, context: dict[str, Any]) -> None:
    # replace with a different github client because we need a different
    # identity to leave the "iambic git-plan". Otherwise, it won't be able
    # to trigger the correct react-to-comment workflow.
    github_client = github.Github(context["iambic"]["GH_OVERRIDE_TOKEN"])
    # repo_name is already in the format {repo_owner}/{repo_short_name}
    repo_name = context["repository"]
    pull_number = context["event"]["pull_request"]["number"]
    # repository_url_token
    templates_repo = github_client.get_repo(repo_name)
    pull_request = templates_repo.get_pull(pull_number)

    try:
        pull_request.create_issue_comment("iambic git-plan")
    except Exception as e:
        captured_traceback = traceback.format_exc()
        log.error("fault", exception=captured_traceback)
        pull_request.create_issue_comment(
            "exception during pull-request is {0} \n ```{1}```".format(
                pull_request.mergeable_state, captured_traceback
            )
        )
        raise e


def handle_detect_changes_from_eventbridge(
    github_client: github.Github, context: dict[str, Any]
) -> None:
    # we need a different github token because we will need to push to main without PR
    github_token = context["iambic"]["GH_OVERRIDE_TOKEN"]
    repository_url = context["event"]["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)

    repo_name = context["repository"]
    templates_repo = github_client.get_repo(repo_name)
    default_branch = get_remote_default_branch(templates_repo)
    _handle_detect_changes_from_eventbridge(repo_url, default_branch)


def _handle_detect_changes_from_eventbridge(
    repo_url: str, default_branch: str
) -> list[TemplateChangeDetails]:
    try:
        repo = prepare_local_repo_for_new_commits(
            repo_url, get_lambda_repo_path(), "detect"
        )

        run_detect(get_lambda_repo_path())
        repo.git.add(".")
        diff_list = repo.head.commit.diff()
        if len(diff_list) > 0:
            repo.git.commit("-m", COMMIT_MESSAGE_FOR_DETECT)
            repo.remotes.origin.push(refspec=f"HEAD:{default_branch}").raise_if_error()
        else:
            log.info("handle_detect no changes")
    except Exception as e:
        log.error("fault", exception=str(e))
        raise e
    return []


def handle_import(github_client: github.Github, context: dict[str, Any]) -> None:
    # we need a different github token because we will need to push to main without PR
    github_token = context["iambic"]["GH_OVERRIDE_TOKEN"]

    # repo_name is already in the format {repo_owner}/{repo_short_name}
    repository_url = context["event"]["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)
    repo_name = context["repository"]
    templates_repo = github_client.get_repo(repo_name)
    default_branch = get_remote_default_branch(templates_repo)
    _handle_import(repo_url, default_branch)


def _handle_import(repo_url: str, default_branch: str) -> list[TemplateChangeDetails]:
    try:
        exe_message = ExecutionMessage(
            execution_id=str(uuid.uuid4()), command=Command.IMPORT
        )
        repo_dir = get_lambda_repo_path()
        repo = prepare_local_repo_for_new_commits(repo_url, repo_dir, "import")
        config_path = asyncio.run(resolve_config_template_path(repo_dir))
        config = asyncio.run(load_config(config_path))
        asyncio.run(config.run_import(exe_message, repo_dir))
        repo.git.add(".")
        diff_list = repo.head.commit.diff()
        if len(diff_list) > 0:
            repo.git.commit("-m", COMMIT_MESSAGE_FOR_IMPORT)
            repo.remotes.origin.push(refspec=f"HEAD:{default_branch}").raise_if_error()
        else:
            log.info("_handle_import no changes")
    except Exception as e:
        log.error("fault", exception=str(e))
        raise e
    return []


def _handle_enforce(repo_url: str, default_branch: str) -> list[TemplateChangeDetails]:
    try:
        local_repo_path = get_lambda_repo_path()
        _ = prepare_local_repo_for_new_commits(repo_url, local_repo_path, "enforce")
        config_path = asyncio.run(resolve_config_template_path(local_repo_path))
        config = asyncio.run(load_config(config_path))
        # we are not restoring teh original ctx because we expect
        # this is called in a completely separate process
        ctx.eval_only = False
        # running in enforce option mean we are only writing to the cloud.
        # there will be no templates being changed in the git context
        template_changes = run_apply(
            config,
            [],
            repo_dir=local_repo_path,
            enforced_only=True,
            output_path=getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH,
        )
        return template_changes
    except Exception as e:
        log.error("fault", exception=str(e))
        raise e


def handle_expire(github_client: github.Github, context: dict[str, Any]) -> None:
    # we need a different github token because we will need to push to main without PR
    github_token = context["iambic"]["GH_OVERRIDE_TOKEN"]

    # repo_name is already in the format {repo_owner}/{repo_short_name}
    repository_url = context["event"]["repository"]["clone_url"]
    repo_url = format_github_url(repository_url, github_token)
    repo_name = context["repository"]
    templates_repo = github_client.get_repo(repo_name)
    default_branch = get_remote_default_branch(templates_repo)
    _handle_expire(repo_url, default_branch)


def _handle_expire(repo_url: str, default_branch: str) -> list[TemplateChangeDetails]:
    try:
        repo = prepare_local_repo_for_new_commits(
            repo_url, get_lambda_repo_path(), "expire"
        )

        run_expire(None, get_lambda_repo_path())
        repo.git.add(".")
        diff_list = repo.head.commit.diff()
        if len(diff_list) > 0:
            repo.git.commit("-m", "expire")  # FIXME

            lambda_run_handler(None, {"command": "git_apply"})

            # if it's in a PR, it's more natural to upload the proposed_changes.yaml to somewhere
            # current implementation, it's just logging to standard out
            lines = []
            filepath = getattr(iambic_app, "lambda").app.PLAN_OUTPUT_PATH
            if os.path.exists(filepath):
                with open(filepath) as f:
                    lines = f.readlines()
            log_params = {"proposed_changes": lines}
            log.info("handle_expire ran", **log_params)

            default_branch = get_remote_default_branch(repo)
            repo.remotes.origin.push(
                refspec=f"HEAD:{default_branch}"
            ).raise_if_error()  # FIXME
        else:
            log.info("handle_expire no changes")
    except Exception as e:
        log.error("fault", exception=str(e))
        raise e
    return []


EVENT_DISPATCH_MAP: dict[str, Callable] = {
    "issue_comment": handle_issue_comment,
    "pull_request": handle_pull_request,
    "iambic_command": handle_iambic_command,
}


IAMBIC_CLOUD_IMPORT_DISPATCH_MAP: dict[str, Callable] = {
    "detect": handle_detect_changes_from_eventbridge,
    "import": handle_import,
    "expire": handle_expire,
}


COMMENT_DISPATCH_MAP: dict[str, Callable] = {
    "iambic git-apply": handle_iambic_git_apply,
    "iambic git-plan": handle_iambic_git_plan,
    "iambic apply": handle_iambic_git_apply,
    "iambic plan": handle_iambic_git_plan,
}

if __name__ == "__main__":
    github_context_json_str = os.environ.get("GITHUB_CONTEXT")
    github_override_token = os.environ.get("GH_OVERRIDE_TOKEN")
    iambic_integration_str = os.environ.get("IAMBIC_CLOUD_IMPORT_CMD")
    iambic_commit_email = os.environ.get("IAMBIC_COMMIT_EMAIL")
    iambic_commit_username = os.environ.get("IAMBIC_COMMIT_USERNAME")
    iambic_commit_message = os.environ.get("IAMBIC_COMMIT_MESSAGE")
    iambic_config_file = os.environ.get("IAMBIC_CONFIG_FILE")

    with open("/root/github_context/github_context.json", "r") as f:
        github_context = json.load(f)
        github_context["iambic"] = {}
        github_context["iambic"]["GH_OVERRIDE_TOKEN"] = github_override_token
        github_context["iambic"]["IAMBIC_CLOUD_IMPORT_CMD"] = iambic_integration_str
        github_context["iambic"]["IAMBIC_COMMIT_EMAIL"] = iambic_commit_email
        github_context["iambic"]["IAMBIC_COMMIT_USERNAME"] = iambic_commit_username
        github_context["iambic"]["IAMBIC_COMMIT_MESSAGE"] = iambic_commit_message
        github_context["iambic"]["IAMBIC_CONFIG_FILE"] = iambic_config_file

    if iambic_integration_str:
        github_context["event_name"] = "iambic_command"

    run_handler(github_context)



from __future__ import annotations

import asyncio
import json
import os
import pathlib
import sys
import uuid
import warnings
from typing import Optional

import click

from iambic.config.dynamic_config import Config, init_plugins, load_config
from iambic.config.utils import (
    check_and_update_resource_limit,
    resolve_config_template_path,
)
from iambic.config.wizard import ConfigurationWizard
from iambic.core.context import ctx
from iambic.core.git import clone_git_repos
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage, TemplateChangeDetails
from iambic.core.parser import load_templates
from iambic.core.utils import (
    exceptions_in_proposed_changes,
    gather_templates,
    init_writable_directory,
)
from iambic.output.text import (
    file_render_resource_changes,
    screen_render_resource_changes,
)
from iambic.request_handler.expire_resources import flag_expired_resources
from iambic.request_handler.git_apply import apply_git_changes
from iambic.request_handler.git_plan import plan_git_changes

warnings.filterwarnings("ignore", category=FutureWarning, module="botocore.client")

os.environ.setdefault("IAMBIC_REPO_DIR", str(pathlib.Path.cwd()))


def output_proposed_changes(
    template_changes: list[TemplateChangeDetails],
    output_path: str = "proposed_changes.txt",
    exit_on_error: bool = True,
):
    if output_path is None:
        output_path = "proposed_changes.txt"
    if template_changes:
        log.info(f"A detailed summary of changes has been saved to {output_path}")
        file_render_resource_changes(output_path, template_changes)

    json_filepath = pathlib.Path(output_path).with_suffix(".json")
    with open(str(json_filepath), "w") as fp:
        json.dump([template_change.dict() for template_change in template_changes], fp)

    if exceptions_in_proposed_changes([change.dict() for change in template_changes]):
        log.error(
            "Exceptions encountered. Some operations failed. Please read proposed_changes for details."
        )
        if exit_on_error:
            raise SystemExit(1)


@click.group()
@click.version_option(package_name="iambic-core")
def cli():
    ...


@cli.command()
@click.argument(
    "templates",
    required=False,
    envvar="IAMBIC_TEMPLATE_PATH",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
def expire(templates: list[str], repo_dir: str):
    run_expire(templates, repo_dir=repo_dir)


def run_expire(templates: list[str], repo_dir: str = str(pathlib.Path.cwd())):
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    # load_config is required to populate known templates
    asyncio.run(load_config(config_path))

    if not templates:
        templates = asyncio.run(gather_templates(repo_dir))

    asyncio.run(flag_expired_resources(templates))


@cli.command()
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
def detect(repo_dir: str):
    run_detect(repo_dir)


def run_detect(repo_dir: str):
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    config = asyncio.run(load_config(config_path))
    asyncio.run(config.run_detect_changes(repo_dir))


@cli.command()
@click.option(
    "--repo_dir",
    "-d",
    "repo_dir",
    required=True,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo base directory that should contain the templates. Example: ~/iambic/templates",
)
def clone_repos(repo_dir: str):
    run_clone_repos(repo_dir=repo_dir)


def run_clone_repos(repo_dir: str = str(pathlib.Path.cwd())):
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    config = asyncio.run(load_config(config_path))
    asyncio.run(clone_git_repos(config, repo_dir))


@cli.command()
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    show_default=True,
    help="Apply changes without asking for permission",
)
@click.argument(
    "templates",
    required=False,
    envvar="IAMBIC_TEMPLATE_PATH",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--enforced-only",
    "-e",
    is_flag=True,
    help="Apply Iambic templates that are in `enforced` mode",
)
@click.option(
    "--allow-dirty",
    is_flag=True,
    hidden=True,
    help="Allow applying changes from a dirty git repo",
)
@click.option(
    "--from-sha",
    "from_sha",
    hidden=True,
    required=False,
    type=str,
    help="The from_sha to calculate diff",
)
@click.option(
    "--to-sha",
    "to_sha",
    hidden=True,
    required=False,
    type=str,
    help="The to_sha to calculate diff",
)
@click.option(
    "--plan-output",
    "-o",
    "plan_output",
    type=click.Path(exists=True),
    help="The location to output the plan Example: ./proposed_changes.yaml",
)
def apply(
    repo_dir: str,
    force: bool,
    templates: list[str],
    enforced_only: bool,
    allow_dirty: bool,
    from_sha: str,
    to_sha: str,
    plan_output: str,
):
    if from_sha:
        if not to_sha:
            log.error("to_sha is required when from_sha is provided")
            return
        if templates:
            log.error("templates cannot be provided when from_sha is provided")
            return
        run_git_apply(
            allow_dirty,
            from_sha,
            to_sha,
            repo_dir=repo_dir,
            output_path=plan_output,
        )
    else:
        if to_sha or from_sha:
            log.error("to_sha and from_sha are not supported with templates")
            return
        ctx.eval_only = not force
        config_path = asyncio.run(resolve_config_template_path(repo_dir))
        config = asyncio.run(load_config(config_path))
        run_apply(config, templates, repo_dir=repo_dir, enforced_only=enforced_only)


def run_apply(
    config: Config,
    templates: Optional[list[str]],
    repo_dir: str = str(pathlib.Path.cwd()),
    enforced_only: bool = False,
    output_path: str = "proposed_changes.yaml",
) -> list[TemplateChangeDetails]:
    template_changes = []
    exe_message = ExecutionMessage(
        execution_id=str(uuid.uuid4()), command=Command.APPLY
    )
    if not templates:
        if not enforced_only:
            log.error("Please pass in specific templates to apply.")
            return template_changes
        templates = asyncio.run(gather_templates(repo_dir))
    templates = load_templates(templates)
    if enforced_only:
        templates = [t for t in templates if t.iambic_managed == IambicManaged.ENFORCED]
    if not templates:
        log.info("No templates found")
        return template_changes
    asyncio.run(flag_expired_resources([template.file_path for template in templates]))
    template_changes = asyncio.run(config.run_apply(exe_message, templates))
    output_proposed_changes(template_changes, output_path=output_path)

    screen_render_resource_changes(template_changes)

    if ctx.eval_only and template_changes and click.confirm("Proceed?"):
        ctx.eval_only = False
        template_changes = asyncio.run(config.run_apply(exe_message, templates))
    # This was here before, but I don't think it's needed. Leaving it here for now to see if anything breaks.
    # asyncio.run(config.run_detect_changes(repo_dir))
    return template_changes


def run_git_apply(
    allow_dirty: bool,
    from_sha: str,
    to_sha: str,
    repo_dir: str = str(pathlib.Path.cwd()),
    output_path: str = None,
) -> list[TemplateChangeDetails]:
    ctx.eval_only = False
    config_path = asyncio.run(resolve_config_template_path(repo_dir))

    template_changes = asyncio.run(
        apply_git_changes(
            config_path,
            repo_dir,
            allow_dirty=allow_dirty,
            from_sha=from_sha,
            to_sha=to_sha,
        )
    )
    output_proposed_changes(template_changes, output_path, exit_on_error=False)
    screen_render_resource_changes(template_changes)
    return template_changes


@cli.command()
@click.argument(
    "templates",
    required=False,
    envvar="IAMBIC_TEMPLATE_PATH",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--plan-output",
    "-o",
    "plan_output",
    hidden=True,
    type=click.Path(exists=True),
    help="The location to output the plan Example: ./proposed_changes.yaml",
)
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
@click.option(
    "--git-aware",
    is_flag=True,
    hidden=True,
)
def plan(templates: list, plan_output: str, repo_dir: str, git_aware: bool):
    if git_aware:
        run_git_plan(plan_output, repo_dir=repo_dir)
    else:
        if not templates:
            log.error("Invalid arguments", error="templates is a required argument")
            raise sys.exit(1)
        run_plan(templates, repo_dir=repo_dir)


def run_git_plan(
    output_path: str,
    repo_dir: str = str(pathlib.Path.cwd()),
) -> list[TemplateChangeDetails]:
    ctx.eval_only = True
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    config = asyncio.run(load_config(config_path))
    check_and_update_resource_limit(config)
    template_changes = asyncio.run(plan_git_changes(config_path, repo_dir))
    output_proposed_changes(template_changes, output_path, exit_on_error=False)
    screen_render_resource_changes(template_changes)
    return template_changes


def run_plan(templates: list[str], repo_dir: str = str(pathlib.Path.cwd())):
    if not templates:
        templates = asyncio.run(gather_templates(repo_dir))

    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    config = asyncio.run(load_config(config_path))
    exe_message = ExecutionMessage(
        execution_id=str(uuid.uuid4()), command=Command.APPLY
    )
    asyncio.run(flag_expired_resources(templates))
    ctx.eval_only = True
    template_changes = asyncio.run(
        config.run_apply(exe_message, load_templates(templates))
    )
    output_proposed_changes(template_changes)
    screen_render_resource_changes(template_changes)


@cli.command()
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
def config_discovery(repo_dir: str):
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    config = asyncio.run(load_config(config_path))
    exe_message = ExecutionMessage(
        execution_id=str(uuid.uuid4()), command=Command.CONFIG_DISCOVERY
    )
    asyncio.run(config.run_discover_upstream_config_changes(exe_message, repo_dir))


@cli.command(name="import")
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
def import_(repo_dir: str):
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    config = asyncio.run(load_config(config_path))
    check_and_update_resource_limit(config)
    exe_message = ExecutionMessage(
        execution_id=str(uuid.uuid4()), command=Command.IMPORT
    )
    asyncio.run(config.run_import(exe_message, repo_dir))


@cli.command()
@click.argument(
    "templates",
    required=False,
    envvar="IAMBIC_TEMPLATE_PATH",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory containing the templates. Example: ~/iambic-templates",
)
def lint(templates: list[str], repo_dir: str):
    ctx.eval_only = True
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    asyncio.run(load_config(config_path, configure_plugins=False))

    if not templates:
        templates = asyncio.run(gather_templates(repo_dir))

    templates = load_templates(templates, False)
    log.info("Formatting templates.")
    for template in templates:
        template.write()


@cli.command(name="init")
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory. Example: ~/iambic-templates",
)
def init_plugins_cmd(repo_dir: str):
    config_path = asyncio.run(resolve_config_template_path(repo_dir))
    asyncio.run(init_plugins(config_path))


@cli.command()
@click.option(
    "--repo-dir",
    "-d",
    "repo_dir",
    required=False,
    type=click.Path(exists=True),
    default=os.getenv("IAMBIC_REPO_DIR"),
    help="The repo directory. Example: ~/iambic-templates",
)
def setup(repo_dir: str):
    ctx.command = Command.APPLY
    ConfigurationWizard(repo_dir).run()


if __name__ == "__main__":
    init_writable_directory()
    cli()



from __future__ import annotations

import asyncio
import contextlib
import os
import pathlib
import re
import sys
import tempfile
import typing
from datetime import date, datetime, timezone
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Coroutine, Optional, Union
from urllib.parse import unquote_plus

import aiofiles
from asgiref.sync import sync_to_async
from ruamel.yaml import YAML

from iambic.core import noq_json as json
from iambic.core.aio_utils import gather_limit
from iambic.core.exceptions import RateLimitException
from iambic.core.iambic_enum import IambicManaged
from iambic.core.logger import log

if TYPE_CHECKING:
    from iambic.core.models import ProposedChange


NOQ_TEMPLATE_REGEX = r".*template_type:\n?.*NOQ::"
RATE_LIMIT_STORAGE: dict[str, int] = {}

__WRITABLE_DIRECTORY__ = pathlib.Path.home()


def init_writable_directory() -> None:
    # use during development
    __WRITABLE_DIRECTORY__ = pathlib.Path.home()

    # use by docker image user
    if os.environ.get("IAMBIC_DOCKER_CONTAINER", False):
        log.info("IAMBIC_DOCKER_CONTAINER is set, using /app as writable directory")
        __WRITABLE_DIRECTORY__ = pathlib.Path("/app")

    # use by lambda deployment - the order matters because lambda deployment
    # still use a docker container but still cannot write to anywhere but os temp dir
    if os.environ.get("AWS_LAMBDA_FUNCTION_NAME", False):
        temp_writable_directory = tempfile.mkdtemp(prefix="lambda")
        log.info(
            f"AWS_LAMBDA_FUNCTION_NAME is set, using {temp_writable_directory} as writable directory"
        )
        __WRITABLE_DIRECTORY__ = pathlib.Path(temp_writable_directory)

    this_module = sys.modules[__name__]
    setattr(this_module, "__WRITABLE_DIRECTORY__", __WRITABLE_DIRECTORY__)


def get_writable_directory() -> pathlib.Path:
    return __WRITABLE_DIRECTORY__


def camel_to_snake(str_obj: str) -> str:
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", str_obj).lower()


def camel_to_kebab(str_obj: str) -> str:
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", str_obj).lower()


def snake_to_camelback(str_obj: str) -> str:
    return re.sub(r"_([a-z])", lambda x: x.group(1).upper(), str_obj)


def snake_to_camelcap(str_obj: str) -> str:
    str_obj = camel_to_snake(
        str_obj
    ).title()  # normalize string and add required case convention
    return str_obj.replace("_", "")  # Remove underscores


async def plugin_apply_wrapper(
    apply_awaitable: Coroutine, proposed_changes: list[ProposedChange]
) -> list[ProposedChange]:
    try:
        await apply_awaitable
    except Exception as e:
        for change in proposed_changes:
            change.exceptions_seen.append(str(e))

    return proposed_changes


async def resource_file_upsert(
    file_path: Union[str, pathlib.Path],
    content_as_dict: dict,
    replace_file: bool = False,
):
    """
    Update or create a resource file with the given content.

    This function updates or creates a resource file at the given file path with the given content, which is
    represented as a dictionary. If the file already exists and `replace_file` is False, the function merges
    the existing content with the new content. If `replace_file` is True, the function overwrites the file
    with the new content.

    Args:
    - file_path (Union[str, pathlib.Path]): The file path for the resource file.
    - content_as_dict (dict): The content to be written to the resource file, represented as a dictionary.
    - replace_file (bool, optional): A flag indicating whether to replace the file if it already exists.
        Default is False.

    Returns:
    - None
    """
    if (
        not replace_file
        and os.path.exists(file_path)
        and os.stat(file_path).st_size != 0
    ):
        async with aiofiles.open(file_path, mode="r") as f:
            content_dict = json.loads(await f.read())
            content_as_dict = {**content_dict, **content_as_dict}

    async with aiofiles.open(file_path, mode="w") as f:
        await f.write(json.dumps(content_as_dict, indent=2))


async def file_regex_search(
    file_path: Union[str, Path], re_pattern: str
) -> Union[str, None]:
    async with aiofiles.open(file_path, mode="r") as f:
        file_content = await f.read()
        if re.search(re_pattern, file_content):
            return file_path


async def gather_templates(repo_dir: str, template_type: str = None) -> list[str]:
    repo_dir_path = Path(repo_dir)
    if not repo_dir_path.is_dir():
        raise ValueError(f"{repo_dir_path} is not a directory")

    if template_type and template_type.startswith("NOQ::"):
        # Strip the prefix, so it plays nice with NOQ_TEMPLATE_REGEX
        template_type = template_type.replace("NOQ::", "")

    regex_pattern = (
        rf"{NOQ_TEMPLATE_REGEX}.*{template_type}"
        if template_type
        else NOQ_TEMPLATE_REGEX
    )
    # since multiple glob pattern can potential intersect, we use a set data structure
    # to suppress any duplicate for defensive measure
    # Support both yaml and yml extensions for templates
    file_path_set = set()
    # >>> glob.glob('**/*.txt', recursive=True)
    # ['2.txt', 'sub/3.txt']
    file_path_set.update(repo_dir_path.glob("**/*.yaml"))
    file_path_set.update(repo_dir_path.glob("**/*.yml"))

    file_paths = await gather_limit(
        *[file_regex_search(fp, regex_pattern) for fp in file_path_set],
        limit=int(os.environ.get("IAMBIC_GATHER_TEMPLATES_LIMIT", 10)),
    )
    return [fp for fp in file_paths if fp]


async def aio_wrapper(fnc, *args, **kwargs):
    thread_sensitive = kwargs.pop("thread_sensitive", False)
    return await sync_to_async(fnc, thread_sensitive=thread_sensitive)(*args, **kwargs)


class NoqYaml(YAML):
    def dump(self, data, stream=None, **kw):
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


class NoqSemaphore:
    def __init__(
        self, callback_function: any, batch_size: int, callback_is_async: bool = True
    ):
        """Makes a reusable semaphore that wraps a provided function.
        Useful for batch processing things that could be rate limited.

        Example logs hello there 3 times in quick succession, waits 3 seconds then processes another 3:
            from datetime import datetime

            async def hello_there():
                log.info(f"Hello there - {datetime.utcnow()}")
                await asyncio.sleep(3)

            hello_there_semaphore = NoqSemaphore(hello_there, 3)
            asyncio.run(hello_there_semaphore.process([{} for _ in range(10)]))
        """
        self.limit = asyncio.Semaphore(batch_size)
        self.callback_function = callback_function
        self.callback_is_async = callback_is_async

    async def handle_message(self, **kwargs):
        async with self.limit:
            if self.callback_is_async:
                return await self.callback_function(**kwargs)

            return await aio_wrapper(self.callback_function, **kwargs)

    async def process(self, messages: list[dict], return_exceptions=False):
        return await asyncio.gather(
            *[asyncio.create_task(self.handle_message(**msg)) for msg in messages],
            return_exceptions=return_exceptions,
        )


async def async_batch_processor(
    tasks: list,
    batch_size: int,
    seconds_between_process: Union[int, float, None] = None,
    return_exceptions: bool = False,
) -> list:
    """
    Batches up tasks in an effort to prevent rate limiting
    """
    if len(tasks) <= batch_size:
        return await asyncio.gather(*tasks, return_exceptions=return_exceptions)

    response = []
    for min_elem in range(0, len(tasks), batch_size):
        response.extend(
            await asyncio.gather(
                *tasks[min_elem : min_elem + batch_size],
                return_exceptions=return_exceptions,
            )
        )
        if len(response) == len(tasks):
            return response

        if seconds_between_process:
            await asyncio.sleep(seconds_between_process)

    return response


def un_wrap_json(json_obj: Any) -> Any:
    """Helper function to unwrap nested JSON in the AWS Config resource configuration."""
    # pylint: disable=C0103,W0703,R0911
    # Is this a field that we can safely return?
    if isinstance(json_obj, (type(None), int, bool, float)):  # noqa
        return json_obj
    # Is this a Datetime? Convert it to a string and return it:
    if isinstance(json_obj, datetime):
        return str(json_obj)
    # Is this a Dictionary?
    if isinstance(json_obj, dict):
        decoded = {k: un_wrap_json(v) for k, v in json_obj.items()}
    elif isinstance(json_obj, list):
        decoded = [un_wrap_json(x) for x in json_obj]
        # Yes, try to sort the contents of lists. This is because AWS does not consistently store list ordering for many resource types:
        with contextlib.suppress(Exception):
            sorted_list = sorted(decoded)
            decoded = sorted_list
    else:
        # Try to load the JSON string:
        try:
            # Check if the string starts with a "[" or a "{" (because apparently '123' is a valid JSON)
            for check_field in {
                "{",
                "[",
                '"{',
                '"[',
            }:  # Some of the double-wrapping is really ridiculous
                if json_obj.startswith(check_field):
                    decoded = json.loads(json_obj)
                    # If we loaded this properly, then we need to pass the decoded JSON back in for all the nested stuff:
                    return un_wrap_json(decoded)
            # Check if this string is URL Encoded - if it is, then re-run it through:
            decoded = unquote_plus(json_obj)
            return un_wrap_json(decoded) if decoded != json_obj else json_obj
        except Exception:  # noqa
            return json_obj
    return decoded


def un_wrap_json_and_dump_values(json_obj: Any) -> Any:
    json_obj = un_wrap_json(json_obj)
    for k, v in json_obj.items():
        json_obj[k] = json.dumps(v)
    return json_obj


# lifted from cloudumi's repo common.lib.generic import sort_dict, and modified to support prioritization
def sort_dict(original, prioritize=None):
    """Recursively sorts dictionary keys and dictionary values in alphabetical order,
    with optional prioritization of certain elements.
    """
    if prioritize is None:
        prioritize = [
            "template_type",
            "name",
            "description",
            "included_accounts",
            "excluded_accounts",
        ]
    if isinstance(original, dict):
        # Make a new "ordered" dictionary. No need for Collections in Python 3.7+
        # Sort the keys in the dictionary
        keys = sorted(original.keys())
        # Move the keys in the prioritized list to the front
        keys = [k for k in prioritize if k in keys] + [
            k for k in keys if k not in prioritize
        ]
        d = {k: v for k, v in [(k, original[k]) for k in keys]}
    else:
        d = original
    for k in d:
        if isinstance(d[k], str):
            continue
        if isinstance(d[k], list) and len(d[k]) > 1 and isinstance(d[k][0], str):
            d[k] = sorted(d[k])
        if isinstance(d[k], dict):
            d[k] = sort_dict(d[k], prioritize)
        if isinstance(d[k], list) and len(d[k]) >= 1 and isinstance(d[k][0], dict):
            for i in range(len(d[k])):
                d[k][i] = sort_dict(d[k][i], prioritize)
    return d


def transform_comments(yaml_dict):
    comment_dict = {}
    yaml_dict["metadata_commented_dict"] = comment_dict
    for key, comment in yaml_dict.ca.items.items():
        comment_dict[key] = comment[2].value
        value = yaml_dict[key]
        if isinstance(value, list) and len(value) > 0 and isinstance(value[0], dict):
            yaml_dict[key] = [transform_comments(n) for n in value]
        elif isinstance(value, dict):
            yaml_dict[key] = transform_comments(value)
    return yaml_dict


def create_commented_map(_dict: dict):
    from ruamel.yaml import CommentedMap

    commented_map = CommentedMap()
    index = 0
    comment_key_to_comment = _dict.pop("metadata_commented_dict", {})
    for key, value in _dict.items():
        if isinstance(value, list) and len(value) > 0 and isinstance(value[0], dict):
            value = [create_commented_map(n) for n in value]
        elif isinstance(value, dict):
            value = create_commented_map(value)
        commented_map.insert(index, key, value, comment_key_to_comment.get(key, None))
        index = index + 1
    return commented_map


typ = "rt"
yaml = NoqYaml(typ=typ)
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.representer.ignore_aliases = lambda *data: True
yaml.width = 4096


def evaluate_on_provider(
    resource,
    provider_details,
    exclude_import_only: bool = True,
) -> bool:
    """
    Determine if the provided resource is on or should be on the given provider.

    This function takes a resource, provider details, and an execution context, and returns a Boolean indicating
    whether the resource should be evaluated on the provider. The evaluation is based on the rules defined in
    the resource's `included_children`, `excluded_children`, `included_parents`, and `excluded_parents`
    attributes, as well as the `iambic_managed` attribute of the resource and the provider details.

    Args:
    - resource: The resource to be evaluated.
    - provider_details: The provider details to use for the evaluation.
    - exclude_import_only (bool, optional): A flag indicating whether to exclude resources that are marked as
        import-only. Default is True.

    Returns:
    - A Boolean indicating whether the resource should be evaluated on the provider.
    """
    from iambic.core.models import AccessModelMixin

    no_op_values = [IambicManaged.DISABLED]
    if exclude_import_only:
        no_op_values.append(IambicManaged.IMPORT_ONLY)

    if (
        provider_details.iambic_managed in no_op_values
        or getattr(resource, "iambic_managed", None) in no_op_values
    ):
        return False

    if not isinstance(resource, AccessModelMixin):
        return True

    if provider_details.parent_id:
        if provider_details.parent_id in resource.excluded_parents:
            return False
        elif "*" not in resource.included_parents and not any(
            re.match(parent_id, provider_details.parent_id)
            for parent_id in resource.included_parents
        ):
            return False

    if not resource.included_children:
        return True

    provider_ids = sorted(provider_details.all_identifiers, key=len, reverse=True)
    included_children = sorted(
        [rule.lower() for rule in resource.included_children], key=len, reverse=True
    )
    excluded_children = sorted(
        [rule.lower() for rule in resource.excluded_children], key=len, reverse=True
    )
    exclude_weight = 0

    for exclude_rule in excluded_children:
        if exclude_rule == "*" or any(
            is_regex_match(exclude_rule, provider_id) for provider_id in provider_ids
        ):
            exclude_weight = len(exclude_rule)
            break

    for include_rule in included_children:
        if include_rule == "*" or any(
            is_regex_match(include_rule, provider_id) for provider_id in provider_ids
        ):
            return bool(len(include_rule) > exclude_weight)

    return False


def apply_to_provider(resource, provider_details) -> bool:
    if hasattr(resource, "deleted") and resource.deleted:
        return False

    return evaluate_on_provider(resource, provider_details)


def is_regex_match(regex, test_string):
    regex = regex.lower()
    test_string = test_string.lower()

    if "*" in regex:
        # Normalize user created regex to python regex
        # Example, dev-* to dev-.* to prevent re.match return True for eval on dev
        try:
            sanitized_regex = regex.replace(".*", "*").replace("*", ".*")
            return bool(re.match(sanitized_regex, test_string))
        except re.error:
            return regex == test_string
    else:
        # it is not an actual regex string, just string comparison
        return regex == test_string


def get_provider_value(matching_values: list, identifiers: set[str]):
    """
    Get the provider value that matches the given identifiers.

    This function takes a list of matching values and a set of identifiers, and returns the matching value that
    has the highest priority for the given identifiers. The priority of the matching values is determined based
    on the rules defined in the `included_children` and `excluded_children` attributes of each matching value.

    Args:
    - matching_values (list): A list of matching values to search through.
    - identifiers (set[str]): A set of identifiers to match against the rules.

    Returns:
    - The matching value that has the highest priority for the given identifiers.
    """
    included_account_map = dict()
    included_account_lists = list()

    for matching_val in matching_values:
        for included_account in matching_val.included_children:
            included_account_map[included_account] = matching_val
            included_account_lists.append(included_account)

    for included_account in sorted(included_account_lists, key=len, reverse=True):
        cur_val = included_account_map[included_account]
        included_children = sorted(
            [rule.lower() for rule in cur_val.included_children], key=len, reverse=True
        )
        excluded_children = sorted(
            [rule.lower() for rule in cur_val.excluded_children], key=len, reverse=True
        )
        exclude_weight = 0

        for exclude_rule in excluded_children:
            if exclude_rule == "*" or any(
                is_regex_match(exclude_rule, provider_id) for provider_id in identifiers
            ):
                exclude_weight = len(exclude_rule)
                break

        for include_rule in included_children:
            if include_rule == "*" or any(
                is_regex_match(include_rule, provider_id) for provider_id in identifiers
            ):
                if len(include_rule) > exclude_weight:
                    return cur_val
                else:
                    break


class GlobalRetryController:
    """
    A context manager class that automatically retries a function in case of rate limit exceptions.

    This class provides a convenient way to wrap a function execution with rate limit handling. It retries the function
    in case of rate limit exceptions and stores the state of the rate limit in a global storage to ensure that all
    similarly executed functions are paused when a rate limit exception is encountered.

    Attributes:
        wait_time (int): The time to wait before retrying the function in case of rate limit exceptions (default is 60
                        seconds).
        retry_exceptions (list): A list of exceptions that will trigger a retry (default is
                                      [TimeoutError, asyncio.exceptions.TimeoutError, RateLimitException]).
        fn_identifier (str): An identifier for the function that is being executed, used to store the state of the rate
                             limit in the global storage (default is None, in which case the function name is used).
        max_retries (int): The maximum number of times to retry the function in case of rate limit exceptions
                           (default is 10).
    """

    def __init__(
        self,
        wait_time: int = 60,
        retry_exceptions: Optional[list[Any]] = None,
        fn_identifier: Optional[str] = None,
        max_retries: int = 10,
    ):
        if retry_exceptions is None:
            retry_exceptions = [
                TimeoutError,
                asyncio.exceptions.TimeoutError,
                RateLimitException,
            ]
        self.wait_time = wait_time
        self.retry_exceptions = retry_exceptions
        self.fn_identifier = fn_identifier
        self.max_retries = max_retries

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __call__(self, func: typing.Callable, *args, **kwargs):
        if self.fn_identifier is not None:
            endpoint = self.fn_identifier
        else:
            endpoint = func.__name__
        retries = 0
        while retries < self.max_retries:
            try:
                res = await func(*args, **kwargs)
                if retries > 0:
                    log.info(f"Retry successful for {endpoint}.")
                return res
            except Exception as e:
                if type(e) not in self.retry_exceptions:
                    raise e
                if self.max_retries == retries + 1:
                    raise e
                if RATE_LIMIT_STORAGE.get(endpoint):
                    wait_time = (
                        RATE_LIMIT_STORAGE[endpoint] - asyncio.get_running_loop().time()
                    )
                    if wait_time > 0:
                        await asyncio.sleep(wait_time)
                    else:
                        del RATE_LIMIT_STORAGE[endpoint]
                RATE_LIMIT_STORAGE[endpoint] = (
                    asyncio.get_running_loop().time() + self.wait_time
                )
                retries += 1
                log.warning(
                    f"Rate limit hit for {endpoint}. Retrying in {self.wait_time} seconds."
                )


def sanitize_string(unsanitized_str, valid_characters_re):
    """
    This function sanitizes the session name typically passed as a parameter name, to ensure it is valid.
    """

    sanitized_str = ""
    max_length = 64  # Session names have a length limit of 64 characters
    for char in unsanitized_str:
        if len(sanitized_str) == max_length:
            break
        if re.match(valid_characters_re, char):
            sanitized_str += char
    return sanitized_str


def simplify_dt(_dt: Union[datetime, date]) -> str:
    """
    Simplify a datetime object  by converting it into a string
    with minute granularity. Useful for handling
    `expires_at` without needing to expose microsecond granularity in our templates.
    """
    if not isinstance(_dt, datetime) and not isinstance(_dt, date):
        return _dt
    dt_str = f"{_dt:%Y-%m-%dT%H:%M:%S}"[:-3]
    if isinstance(_dt, datetime) and _dt.tzinfo:
        dt_str += f" {_dt:%Z}"
    else:
        dt_str += " UTC"
    return dt_str


def normalize_dict_keys(
    obj, case_convention=camel_to_snake, skip_formatting_keys: list = None
):
    if not skip_formatting_keys:
        skip_formatting_keys = ["condition"]
    if isinstance(obj, dict):
        new_obj = dict()
        for k, v in obj.items():
            k = case_convention(k)
            if isinstance(v, list):
                new_obj[k] = [
                    normalize_dict_keys(x, case_convention, skip_formatting_keys)
                    for x in v
                ]
            else:
                new_obj[k] = (
                    normalize_dict_keys(v, case_convention, skip_formatting_keys)
                    if k not in skip_formatting_keys
                    else v
                )
        return new_obj
    elif isinstance(obj, list):
        return [
            normalize_dict_keys(x, case_convention, skip_formatting_keys) for x in obj
        ]
    else:
        return obj


def exceptions_in_proposed_changes(obj) -> bool:
    if isinstance(obj, dict):
        if obj.get("exceptions_seen"):
            return True

        new_obj = dict()
        for k, v in obj.items():
            if isinstance(v, list):
                new_obj[k] = any(exceptions_in_proposed_changes(x) for x in v)
            elif isinstance(v, dict):
                new_obj[k] = exceptions_in_proposed_changes(v)
        return any(list(new_obj.values())) if new_obj else False
    elif isinstance(obj, list):
        return any(exceptions_in_proposed_changes(x) for x in obj)


async def remove_expired_resources(
    resource,
    template_resource_type: str,
    template_resource_id: str,
    delete_resource_if_expired: bool = True,
):
    from iambic.core.models import BaseModel

    if not isinstance(resource, BaseModel):
        return resource

    log_params = {}
    if hasattr(resource, "resource_type"):
        log_params["resource_type"] = resource.resource_type
    if hasattr(resource, "resource_id"):
        log_params["resource_id"] = resource.resource_id
    if template_resource_type != log_params.get(
        "resource_type"
    ) or template_resource_id != log_params.get("resource_id"):
        log_params["parent_resource_type"] = template_resource_type
        log_params["parent_resource_id"] = template_resource_id

    if isinstance(resource, BaseModel) and hasattr(resource, "expires_at"):
        if resource.expires_at:
            cur_time = datetime.now(tz=timezone.utc)
            if resource.expires_at < cur_time:
                log.info("Expired resource found, marking for deletion", **log_params)
                resource.deleted = True
                return resource

    for field_name in resource.__fields__.keys():
        field_val = getattr(resource, field_name)
        if isinstance(field_val, list):
            new_value = await asyncio.gather(
                *[
                    remove_expired_resources(
                        elem, template_resource_type, template_resource_id
                    )
                    for elem in field_val
                ]
            )
            setattr(resource, field_name, new_value)
            if delete_resource_if_expired:
                for elem in new_value:
                    if getattr(elem, "deleted", None) is True:
                        new_value.remove(elem)
                        setattr(resource, field_name, new_value)
        else:
            new_value = await remove_expired_resources(
                field_val, template_resource_type, template_resource_id
            )
            if getattr(new_value, "deleted", None) is True:
                setattr(resource, field_name, None)
            else:
                setattr(resource, field_name, new_value)

    return resource



from __future__ import annotations

from iambic.core.logger import log


class BaseException(Exception):
    def __init__(self, msg: str = "") -> None:
        self.msg = msg
        log.error("An error occurred", error=msg, exception=self.__class__.__name__)
        super().__init__(msg)

    def __str__(self):
        """Stringifies the message."""
        return self.msg


class RateLimitException(BaseException):
    """Rate Limit Exception"""


class MultipleSecretsNotAcceptedException(BaseException):
    def __init__(self):
        super().__init__("extends tag does not accept multiples secrets")



from __future__ import annotations

import asyncio
import re
from enum import Enum
from typing import TYPE_CHECKING, Optional, Union

import boto3
from botocore.exceptions import ClientError, NoCredentialsError

from iambic.core.iambic_enum import IambicManaged
from iambic.core.logger import log
from iambic.core.utils import aio_wrapper

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig


def calculate_import_preference(existing_template):
    prefer_templatized = False
    try:
        if existing_template:
            # this is expensive just to compute if
            # the existing template is already bias toward templatized.
            existing_template_in_str = existing_template.json()
            prefer_templatized = "{{" in existing_template_in_str
    except Exception as exc_info:
        # We are willing to tolerate exception because
        # we are calculating preference from possibly bad templates on disk
        log_params = {"exc_info": str(exc_info)}
        log.error("cannot calculate preference from existing template", **log_params)
    return prefer_templatized


async def boto_crud_call(
    boto_fnc, retryable_errors: list = None, **kwargs
) -> Union[list, dict]:
    """Responsible for calls to boto. Adds async support and error handling
    :param boto_fnc:
    :param retryable_errors: A list of error codes that should be retried
    :param kwargs: The params to pass to the boto fnc
    :return:
    """
    retry_count = 0
    if not retryable_errors:
        retryable_errors = []

    retryable_errors.append("Throttling")

    while True:
        try:
            return await aio_wrapper(boto_fnc, **kwargs)
        except ClientError as err:
            if any(
                retryable_err in err.response["Error"]["Code"]
                for retryable_err in retryable_errors
            ):
                if retry_count >= 10:
                    raise
                retry_count += 1
                log.warning(f"Throttling error on {str(boto_fnc)}")
                await asyncio.sleep(retry_count / 4)
                continue
            elif "AccessDenied" in err.response["Error"]["Code"]:
                raise
            else:
                raise
        except NoCredentialsError as exc:
            log.error(
                f"Unable to create an AWS session, you may need to run `aws configure` or export your AWS_PROFILE; err={exc}"
            )
            raise RuntimeError(
                f"Unable to create an AWS session, you may need to run `aws configure` or export your AWS_PROFILE; err={exc}"
            )


async def paginated_search(
    search_fnc,
    response_key: str = None,
    response_keys: list[str] = None,
    max_results: int = None,
    retain_key: bool = False,
    **search_kwargs,
) -> Union[list, dict]:
    """Retrieve and aggregate each paged response, returning a single list of each response object
    :param search_fnc:
    :param response_key:
    :param response_keys:
    :param max_results:
    :param retain_key: If true, the response_key will be retained in the response
    :return:
    """
    assert bool(response_key) ^ bool(response_keys)  # XOR

    if response_keys and not retain_key:
        log.warning("retain_key must be true if response_keys is provided")
        retain_key = True

    if response_key:
        response_keys = [response_key]

    results = {key: [] for key in response_keys}

    while True:
        response = await boto_crud_call(search_fnc, **search_kwargs)
        for response_key in response_keys:
            results[response_key].extend(response.get(response_key, []))

        if not response["IsTruncated"] or (max_results and len(results) >= max_results):
            return results if retain_key else results[response_key]
        else:
            search_kwargs["Marker"] = response["Marker"]


def get_identity_arn(caller_identity: dict) -> str:
    arn = caller_identity.get("Arn")
    arn_split_by_colon = arn.split(":")
    resource_path_split_by_slash = []
    maybe_assumed_role = ""
    role_name = ""

    if len(arn_split_by_colon) < 6:
        # magic number 6 is the expected number of parts for sts caller identity arn
        # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference-arns.html
        return arn

    resource_path_split_by_slash = arn_split_by_colon[5].split("/")

    if len(resource_path_split_by_slash) < 2:
        # magic number 2 is from the format of assumed-role/role-name/role-session-name format
        return arn

    maybe_assumed_role = resource_path_split_by_slash[0]
    role_name = resource_path_split_by_slash[1]

    if maybe_assumed_role != "assumed-role":
        # for example, user-like arn
        return arn

    if maybe_assumed_role == "assumed-role" and role_name.startswith("AWSReservedSSO_"):
        return arn

    # We attempt to derived IAM role principals from Assumed-role session principals
    # AWS Identity Center generated roles can have a path that is under path
    # aws-reserved/sso.amazonaws.com/us-west-2/
    # Path information is not contained within the caller_identity dictionary
    # In the event if we see prefix with AWSReservedSSO_, we fall back to
    # assumed-role session principals
    # See https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_principal.html#Principal_specifying
    identity_arn_with_session_name = arn.replace(":sts:", ":iam:").replace(
        "assumed-role", "role"
    )
    return "/".join(identity_arn_with_session_name.split("/")[:-1])


def get_current_role_arn(sts_client) -> str:
    return get_identity_arn(sts_client.get_caller_identity())


async def legacy_paginated_search(
    search_fnc,
    response_key: str,
    max_results: int = None,
    retain_key: bool = False,
    **search_kwargs,
) -> list:
    """Retrieve and aggregate each paged response, returning a single list of each response object

    Why is there 2 paginated searches? - AWS has 2 ways of paginating results
    This one seems to be the older way which is why it's called legacy

    :param search_fnc:
    :param response_key:
    :param max_results:
    :param retain_key: If true, the response_key will be retained in the response
    :return:
    """
    results = []
    is_first_call = True

    while True:
        response = await boto_crud_call(search_fnc, **search_kwargs)
        results.extend(response.get(response_key, []))

        if (
            not response.get("NextToken")
            or (max_results and len(results) >= max_results)
            or (not results and not is_first_call)
        ):
            return {response_key: results} if retain_key else results
        else:
            is_first_call = False
            search_kwargs["NextToken"] = response["NextToken"]


class RegionName(Enum):
    us_east_1 = "us-east-1"
    us_west_1 = "us-west-1"
    us_west_2 = "us-west-2"
    eu_west_1 = "eu-west-1"
    eu_west_2 = "eu-west-2"
    eu_central_1 = "eu-central-1"
    ap_southeast_1 = "ap-southeast-1"
    ap_southeast_2 = "ap-southeast-2"
    ap_northeast_1 = "ap-northeast-1"
    ap_northeast_2 = "ap-northeast-2"
    sa_east_1 = "sa-east-1"
    # cn_north_1 = "cn-north-1"


def is_valid_account_id(account_id: str):
    return bool(re.match(r"^\d{12}$", account_id))


async def set_org_account_variables(client, account: dict) -> dict:
    tags = await legacy_paginated_search(
        client.list_tags_for_resource, "Tags", ResourceId=account["Id"]
    )
    account["variables"] = [
        {"key": tag["Key"], "value": tag.get("Value")} for tag in tags
    ]
    return account


async def get_aws_account_map(config: AWSConfig) -> dict:
    """Returns a map containing all enabled account configs across all provided config instances

    :param config:
    :return: dict(account_id:str = AWSAccount)
    """
    aws_account_map = dict()
    for aws_account in config.accounts:
        if aws_account_map.get(aws_account.account_id):
            log.critical(
                "Account definition found in multiple configs",
                account_id=aws_account.account_id,
                account_name=aws_account.account_name,
            )
            raise ValueError
        elif aws_account.iambic_managed == IambicManaged.DISABLED:
            log.info(
                "IAMbic awareness disabled for the account. Skipping.",
                account_id=aws_account.account_id,
            )
            continue

        aws_account_map[aws_account.account_id] = aws_account

    return aws_account_map


async def create_assume_role_session(
    boto3_session,
    assume_role_arn: str,
    region_name: str,
    external_id: Optional[str] = None,
    session_name: str = "iambic",
) -> boto3.Session:
    if session_name is None:
        session_name = "iambic"
    try:
        sts = boto3_session.client(
            "sts",
            endpoint_url=f"https://sts.{region_name}.amazonaws.com",
            region_name=region_name,
        )
        role_params = dict(RoleArn=assume_role_arn, RoleSessionName=session_name)
        if external_id:
            role_params["ExternalId"] = external_id
        role = await aio_wrapper(sts.assume_role, **role_params)
        return boto3.Session(
            region_name=region_name,
            aws_access_key_id=role["Credentials"]["AccessKeyId"],
            aws_secret_access_key=role["Credentials"]["SecretAccessKey"],
            aws_session_token=role["Credentials"]["SessionToken"],
        )
    except Exception as err:
        log.error("Failed to assume role", assume_role_arn=assume_role_arn, error=err)
        raise


def boto3_retry(f):
    async def wrapper(*args, **kwargs):
        max_retries = kwargs.pop("max_retries", 10)
        for attempt in range(max_retries):
            try:
                return await f(*args, **kwargs)
            except ClientError as err:
                if (
                    err.response["Error"]["Code"] == "Throttling"
                    and attempt < max_retries - 1
                ):
                    await asyncio.sleep(1)
                else:
                    raise

    return wrapper



from __future__ import annotations

from typing import List

import rich

from iambic.core.models import TemplateChangeDetails
from iambic.output import get_template_env
from iambic.output.models import get_template_data


def file_render_resource_changes(
    filepath: str,
    resource_changes: List[TemplateChangeDetails],
) -> str:
    # Get template data
    template_data = get_template_data(resource_changes)
    # Get template environment
    env = get_template_env()
    # Render template
    template = env.get_template("text_file_summary.jinja2")
    rendered_data = template.render(iambic=template_data)
    with open(filepath, "w") as f:
        f.write(rendered_data)


def screen_render_resource_changes(resource_changes: List[TemplateChangeDetails]):
    template_data = get_template_data(resource_changes)
    env = get_template_env()
    template = env.get_template("text_screen_summary.jinja2")
    rendered_template = template.render(iambic=template_data)
    rich.print(rendered_template)
    return rendered_template
